"""
A place for Datasets and Dataloaders.
"""

from collections import defaultdict
from functools import cache
from pathlib import Path
from typing import Callable, Union

import numpy as np
from skimage.transform import resize

from .config import MicroSplitDataConfig
from .types import DataSplitType, TilingMode
from .utils.empty_patch_fetcher import EmptyPatchFetcher
from .utils.index_manager import GridIndexManagerRef


class MultiChDloaderRef:
    def __init__(
        self,
        data_config: MicroSplitDataConfig,
        fpath: str,
        load_data_fn: Callable,
        val_fraction: float = None,
        test_fraction: float = None,
    ):
        """ """
        self._data_type = data_config.data_type
        self._fpath = Path(fpath)
        self._data = None
        self._3Ddata = False  # TODO wtf it was 5D
        self._tiling_mode = data_config.tiling_mode
        # by default, if the noise is present, add it to the input and target.
        self._depth3D = data_config.depth3D
        self._mode_3D = data_config.mode_3D
        # NOTE: Input is the sum of the different channels. It is not the average of the different channels.
        self._input_is_sum = data_config.input_is_sum
        self._num_channels = data_config.num_channels
        self._input_idx = data_config.input_idx
        self._tar_idx_list = data_config.target_idx_list

        self.load_data(
            data_config,
            data_config.datasplit_type,
            load_data_fn=load_data_fn,
            val_fraction=val_fraction,
            test_fraction=test_fraction,
            allow_generation=data_config.allow_generation,
        )

        self._data_shapes = self.get_data_shapes()
        self._normalized_input = data_config.normalized_input
        self._quantile = 1.0
        self._channelwise_quantile = False
        self._background_quantile = 0.0
        self._clip_background_noise_to_zero = False
        self._skip_normalization_using_mean = False
        self._empty_patch_replacement_enabled = False

        self._background_values = None

        self._overlapping_padding_kwargs = data_config.overlapping_padding_kwargs
        if self._tiling_mode in [TilingMode.TrimBoundary, TilingMode.ShiftBoundary]:
            if (
                self._overlapping_padding_kwargs is None
                or data_config.multiscale_lowres_count is not None
            ):
                # raise warning
                print("Padding is not used with this alignement style")
        else:
            assert (
                self._overlapping_padding_kwargs is not None
            ), "When not trimming boudnary, padding is needed."

        self._is_train = data_config.datasplit_type == DataSplitType.Train

        # input = alpha * ch1 + (1-alpha)*ch2.
        # alpha is sampled randomly between these two extremes
        self._start_alpha_arr = self._end_alpha_arr = self._return_alpha = None

        self._img_sz = self._grid_sz = self._repeat_factor = self.idx_manager = None

        # changed set_img_sz because "grid_size" in data_config returns false
        try:
            grid_size = data_config.grid_size
        except AttributeError:
            grid_size = data_config.image_size

        if self._is_train:
            self._start_alpha_arr = data_config.start_alpha  # TODO why only for train?
            self._end_alpha_arr = data_config.end_alpha

        self.set_img_sz(data_config.image_size, grid_size)

        self._empty_patch_replacement_enabled = (
            data_config.empty_patch_replacement_enabled and self._is_train
        )
        if self._empty_patch_replacement_enabled:
            self._empty_patch_replacement_channel_idx = (
                data_config.empty_patch_replacement_channel_idx
            )
            self._empty_patch_replacement_probab = (
                data_config.empty_patch_replacement_probab
            )
            data_frames = self._data[..., self._empty_patch_replacement_channel_idx]
            # NOTE: This is on the raw data. So, it must be called before removing the background.
            self._empty_patch_fetcher = EmptyPatchFetcher(
                self.idx_manager,
                self._img_sz,
                data_frames,
                max_val_threshold=data_config.empty_patch_max_val_threshold,
            )

        self.rm_bkground_set_max_val_and_upperclip_data(
            data_config.max_val, data_config.datasplit_type
        )

        # For overlapping dloader, image_size and repeat_factors are not related. hence a different function.

        self._mean = None
        self._std = None
        self._use_one_mu_std = data_config.use_one_mu_std

        self._target_separate_normalization = data_config.target_separate_normalization

        self._enable_rotation = data_config.enable_rotation_aug
        flipz_3D = data_config.random_flip_z_3D
        self._flipz_3D = flipz_3D and self._enable_rotation

        self._enable_random_cropping = data_config.enable_random_cropping
        self._uncorrelated_channels = (
            data_config.uncorrelated_channels and self._is_train
        )
        self._uncorrelated_channel_probab = data_config.uncorrelated_channel_probab
        assert self._is_train or self._uncorrelated_channels is False
        assert (
            self._enable_random_cropping is True or self._uncorrelated_channels is False
        )
        # Randomly rotate [-90,90]

        self._rotation_transform = None
        if self._enable_rotation:
            # TODO: fix this import
            import albumentations as A

            self._rotation_transform = A.Compose([A.Flip(), A.RandomRotate90()])

        # TODO: remove print log messages
        # if print_vars:
        #     msg = self._init_msg()
        #     print(msg)

    def get_data_shapes(self):
        if self._3Ddata:  # TODO we assume images don't have a channel dimension
            [
                [
                    im.shape if len(im.shape) == 4 else (1, *im.shape)
                    for im in self._data[ch]
                ]
                for ch in range(len(self._data))
            ]
        else:
            return [
                [
                    im.shape if len(im.shape) == 3 else (1, *im.shape)
                    for im in self._data[ch]
                ]
                for ch in range(len(self._data))
            ]

    def load_data(
        self,
        data_config: MicroSplitDataConfig,
        datasplit_type: DataSplitType,
        load_data_fn: Callable,
        val_fraction=None,
        test_fraction=None,
        allow_generation=None,
    ):
        self._data = load_data_fn(
            data_config,
            self._fpath,
            datasplit_type,
            val_fraction=val_fraction,
            test_fraction=test_fraction,
            allow_generation=allow_generation,
        )

    # TODO check for 2D/3D data consistency with config
    # TODO check number of channels consistency with config

    def save_background(self, channel_idx, frame_idx, background_value):
        self._background_values[frame_idx, channel_idx] = background_value

    def get_background(self, channel_idx, frame_idx):
        return self._background_values[frame_idx, channel_idx]

    def rm_bkground_set_max_val_and_upperclip_data(self, max_val, datasplit_type):
        # self.remove_background() # TODO revisit
        self.set_max_val(max_val, datasplit_type)
        self.upperclip_data()

    def upperclip_data(self):
        for ch_idx, data in enumerate(self._data):
            if self.max_val[ch_idx] is not None:
                for idx in range(len(data)):
                    data[idx][data[idx] > self.max_val[ch_idx]] = self.max_val[ch_idx]

    def compute_max_val(self):
        # TODO add channelwise quantile ?
        return [
            max([np.quantile(im, self._quantile) for im in ch]) for ch in self._data
        ]

    def set_max_val(self, max_val, datasplit_type):
        if max_val is None:
            assert datasplit_type in [DataSplitType.Train, DataSplitType.All]
            self.max_val = self.compute_max_val()
        else:
            assert max_val is not None
            self.max_val = max_val

    def get_max_val(self):
        return self.max_val

    def get_img_sz(self):
        return self._img_sz

    def get_num_frames(self):
        """Returns the number of the longest channel."""
        return max(self.idx_manager.total_grid_count()[0])

    def reduce_data(
        self,
        t_list=None,
        z_start=None,
        z_end=None,
        h_start=None,
        h_end=None,
        w_start=None,
        w_end=None,
    ):
        raise NotImplementedError("Not implemented")

    def get_idx_manager_shapes(
        self, patch_size: int, grid_size: Union[int, tuple[int, int, int]]
    ):
        numC = len(self._data_shapes)
        if self._3Ddata:
            patch_shape = (1, self._depth3D, patch_size, patch_size)
            if isinstance(grid_size, int):
                grid_shape = (1, 1, grid_size, grid_size)
            else:
                assert len(grid_size) == 3
                assert all(
                    [g <= p for g, p in zip(grid_size, patch_shape[1:-1])]
                ), f"Grid size {grid_size} must be less than patch size {patch_shape[1:-1]}"
                grid_shape = (1, grid_size[0], grid_size[1], grid_size[2])
        else:
            assert isinstance(grid_size, int)
            grid_shape = (1, grid_size, grid_size)
            patch_shape = (1, patch_size, patch_size)

        return patch_shape, grid_shape

    def set_img_sz(self, image_size, grid_size: Union[int, tuple[int, int, int]]):
        """
        If one wants to change the image size on the go, then this can be used.
        Args:
            image_size: size of one patch
            grid_size: frame is divided into square grids of this size. A patch centered on a grid having size `image_size` is returned.
        """
        # hacky way to deal with image shape from new conf
        self._img_sz = image_size[-1]  # TODO revisit!
        self._grid_sz = grid_size
        shapes = self._data_shapes

        patch_shape, grid_shape = self.get_idx_manager_shapes(
            self._img_sz, self._grid_sz
        )
        self.idx_manager = GridIndexManagerRef(
            shapes, grid_shape, patch_shape, self._tiling_mode
        )

    def __len__(self):
        # If channel length is not equal, return the longest
        return max(self.idx_manager.total_grid_count()[0])

    def _init_msg(
        self,
    ):
        msg = (
            f"[{self.__class__.__name__}] Train:{int(self._is_train)} Sz:{self._img_sz}"
        )
        dim_sizes = [
            self.idx_manager.get_individual_dim_grid_count(dim)
            for dim in range(len(self._data.shape))
        ]
        dim_sizes = ",".join([str(x) for x in dim_sizes])
        msg += f" N:{self.N} NumPatchPerN:{self._repeat_factor}"
        msg += f"{self.idx_manager.total_grid_count()} DimSz:({dim_sizes})"
        msg += f" TrimB:{self._tiling_mode}"
        # msg += f' NormInp:{self._normalized_input}'
        # msg += f' SingleNorm:{self._use_one_mu_std}'
        msg += f" Rot:{self._enable_rotation}"
        if self._flipz_3D:
            msg += f" FlipZ:{self._flipz_3D}"

        msg += f" RandCrop:{self._enable_random_cropping}"
        msg += f" Channel:{self._num_channels}"
        # msg += f' Q:{self._quantile}'
        if self._input_is_sum:
            msg += f" SummedInput:{self._input_is_sum}"

        if self._empty_patch_replacement_enabled:
            msg += f" ReplaceWithRandSample:{self._empty_patch_replacement_enabled}"
        if self._uncorrelated_channels:
            msg += f" Uncorr:{self._uncorrelated_channels}"
        if self._empty_patch_replacement_enabled:
            msg += f"-{self._empty_patch_replacement_channel_idx}-{self._empty_patch_replacement_probab}"
        if self._background_quantile > 0.0:
            msg += f" BckQ:{self._background_quantile}"

        if self._start_alpha_arr is not None:
            msg += f" Alpha:[{self._start_alpha_arr},{self._end_alpha_arr}]"
        return msg

    def _crop_imgs(self, ch_idx: int, patch_idx: int, img: np.ndarray):
        h, w = img.shape[-2:]
        if self._img_sz is None:
            return (
                img,
                {"h": [0, h], "w": [0, w], "hflip": False, "wflip": False},
            )

        if self._enable_random_cropping:
            # this parameter is ambiguous. It toggles between random/deterministic patching
            patch_start_loc = self._get_random_hw(h, w)
            if self._3Ddata:
                patch_start_loc = (
                    np.random.choice(1 + img.shape[-3] - self._depth3D),
                ) + patch_start_loc
        else:
            # Patch coordinates are calculated by the index manager.
            patch_start_loc = self._get_deterministic_loc(ch_idx, patch_idx)
        cropped_img = self._crop_flip_img(img, patch_start_loc, False, False)

        return cropped_img

    def _crop_img(self, img: np.ndarray, patch_start_loc: tuple):
        if self._tiling_mode in [TilingMode.TrimBoundary, TilingMode.ShiftBoundary]:
            # In training, this is used.
            # NOTE: It is my opinion that if I just use self._crop_img_with_padding, it will work perfectly fine.
            # The only benefit this if else loop provides is that it makes it easier to see what happens during training.
            patch_end_loc = (
                np.array(patch_start_loc, dtype=np.int32)
                + self.idx_manager.patch_shape[1:-1]
            )
            if self._3Ddata:
                z_start, h_start, w_start = patch_start_loc
                z_end, h_end, w_end = patch_end_loc
                new_img = img[..., z_start:z_end, h_start:h_end, w_start:w_end]
            else:
                h_start, w_start = patch_start_loc
                h_end, w_end = patch_end_loc
                new_img = img[..., h_start:h_end, w_start:w_end]

            return new_img
        else:
            # During evaluation, this is used. In this situation, we can have negative h_start, w_start. Or h_start +self._img_sz can be larger than frame
            # In these situations, we need some sort of padding. This is not needed  in the LeftTop alignement.
            return self._crop_img_with_padding(img, patch_start_loc)

    def get_begin_end_padding(self, start_pos, end_pos, max_len):
        """
        The effect is that the image with size self._grid_sz is in the center of the patch with sufficient
        padding on all four sides so that the final patch size is self._img_sz.
        """
        pad_start = 0
        pad_end = 0
        if start_pos < 0:
            pad_start = -1 * start_pos

        pad_end = max(0, end_pos - max_len)

        return pad_start, pad_end

    def _crop_img_with_padding(
        self, img: np.ndarray, patch_start_loc, max_len_vals=None
    ):
        if max_len_vals is None:
            max_len_vals = self.idx_manager.data_shape[1:-1]
        patch_end_loc = np.array(patch_start_loc, dtype=int) + np.array(
            self.idx_manager.patch_shape[1:-1], dtype=int
        )
        boundary_crossed = []
        valid_slice = []
        padding = [[0, 0]]
        for start_idx, end_idx, max_len in zip(
            patch_start_loc, patch_end_loc, max_len_vals
        ):
            boundary_crossed.append(end_idx > max_len or start_idx < 0)
            valid_slice.append((max(0, start_idx), min(max_len, end_idx)))
            pad = [0, 0]
            if boundary_crossed[-1]:
                pad = self.get_begin_end_padding(start_idx, end_idx, max_len)
            padding.append(pad)
        # max() is needed since h_start could be negative.
        if self._3Ddata:
            new_img = img[
                ...,
                valid_slice[0][0] : valid_slice[0][1],
                valid_slice[1][0] : valid_slice[1][1],
                valid_slice[2][0] : valid_slice[2][1],
            ]
        else:
            new_img = img[
                ...,
                valid_slice[0][0] : valid_slice[0][1],
                valid_slice[1][0] : valid_slice[1][1],
            ]

        # print(np.array(padding).shape, img.shape, new_img.shape)
        # print(padding)
        if not np.all(padding == 0):
            new_img = np.pad(new_img, padding, **self._overlapping_padding_kwargs)

        return new_img

    def _crop_flip_img(
        self, img: np.ndarray, patch_start_loc: tuple, h_flip: bool, w_flip: bool
    ):
        new_img = self._crop_img(img, patch_start_loc)
        if h_flip:
            new_img = new_img[..., ::-1, :]
        if w_flip:
            new_img = new_img[..., :, ::-1]

        return new_img.astype(np.float32)

    def _load_img(self, ch_idx: int, patch_idx: int) -> tuple[np.ndarray, np.ndarray]:
        """
        Returns the channels and also the respective noise channels.
        """
        patch_loc_list = self.idx_manager.get_patch_location_from_patch_idx(
            ch_idx, patch_idx
        )
        # TODO we should be adding channel dim here probably
        img = self._data[ch_idx][patch_loc_list[0]]
        return img

    def get_mean_std(self):
        return self._mean, self._std

    def set_mean_std(self, mean_val, std_val):
        self._mean = mean_val
        self._std = std_val

    def normalize_target(self, target):
        mean_dict, std_dict = self.get_mean_std()
        mean_ = mean_dict["target"]  # .squeeze(0)
        std_ = std_dict["target"]  # .squeeze(0)
        return (target - mean_) / std_

    def get_grid_size(self):
        return self._grid_sz

    def get_idx_manager(self):
        return self.idx_manager

    def per_side_overlap_pixelcount(self):
        return (self._img_sz - self._grid_sz) // 2

    def _get_deterministic_loc(self, ch_idx: int, patch_idx: int):
        """
        It returns the top-left corner of the patch corresponding to index.
        """
        loc_list = self.idx_manager.get_patch_location_from_patch_idx(ch_idx, patch_idx)
        # last dim is channel. we need to take the third and the second last element.
        return loc_list[2:]

    @cache
    def crop_probablities(self, ch_idx):
        sizes = np.array([np.prod(x.shape) for x in self._data[ch_idx]])
        return sizes / sizes.sum()

    def sample_crop(self, ch_idx):
        idx = None
        count = 0
        while idx is None:
            count += 1
            idx = np.random.choice(
                len(self._data[ch_idx]), p=self.crop_probablities(ch_idx)
            )
            data = self._data[ch_idx][idx]  # TODO no channel and S dim ?
            # changed for ndim
            if all(
                d >= self._img_sz for d in data.shape[-2:]
            ):  # TODO dims were hardcoded
                h = np.random.randint(0, data.shape[-2] - self._img_sz)
                w = np.random.randint(0, data.shape[-1] - self._img_sz)

                if len(data.shape) > 2 and not self._3Ddata:
                    s = np.random.randint(0, data.shape[0] - 1)
                    return data[s, h : h + self._img_sz, w : w + self._img_sz]
                else:
                    return data[h : h + self._img_sz, w : w + self._img_sz]

            elif count > 100:
                raise ValueError("Cannot find a valid crop")
            else:
                idx = None

        return None

    def _l2(self, x):
        return np.sqrt(np.mean(np.array(x) ** 2))

    def compute_mean_std(self, allow_for_validation_data=False):
        """
        Note that we must compute this only for training data.
        """
        if self._3Ddata:
            raise NotImplementedError("Not implemented for 3D data")

        if self._input_is_sum:
            mean_tar_dict = defaultdict(list)
            std_tar_dict = defaultdict(list)
            mean_inp = []
            std_inp = []
            for _ in range(30000):
                crops = []
                for ch_idx in range(len(self._data)):
                    crop = self.sample_crop(ch_idx)
                    mean_tar_dict[ch_idx].append(np.mean(crop))
                    std_tar_dict[ch_idx].append(np.std(crop))
                    crops.append(crop)

                inp = 0
                for img in crops:
                    inp += img

                mean_inp.append(np.mean(inp))
                std_inp.append(np.std(inp))

            output_mean = defaultdict(list)
            output_std = defaultdict(list)

            NC = len(self._data)
            for ch_idx in range(NC):
                output_mean["target"].append(np.mean(mean_tar_dict[ch_idx]))
                output_std["target"].append(self._l2(std_tar_dict[ch_idx]))

            output_mean["target"] = np.array(output_mean["target"]).reshape(NC, 1, 1)
            output_std["target"] = np.array(output_std["target"]).reshape(NC, 1, 1)

            output_mean["input"] = np.array([np.mean(mean_inp)]).reshape(1, 1, 1)
            output_std["input"] = np.array([self._l2(std_inp)]).reshape(1, 1, 1)
        else:
            raise NotImplementedError("Not implemented for non-summed input")

        return dict(output_mean), dict(output_std)

    def set_mean_std(self, mean_dict, std_dict):
        self._data_mean = mean_dict
        self._data_std = std_dict

    def get_mean_std(self):
        return self._data_mean, self._data_std

    def _get_random_hw(self, h: int, w: int):
        """
        Random starting position for the crop for the img with index `index`.
        """
        if h != self._img_sz:
            h_start = np.random.choice(h - self._img_sz)
            w_start = np.random.choice(w - self._img_sz)
        else:
            h_start = 0
            w_start = 0
        return h_start, w_start

    def replace_with_empty_patch(self, img_tuples):
        """
        Replaces the content of one of the channels with background
        """
        empty_index = self._empty_patch_fetcher.sample()
        empty_img_tuples, empty_img_noise_tuples = self._get_img(empty_index)
        assert (
            len(empty_img_noise_tuples) == 0
        ), "Noise is not supported with empty patch replacement"
        final_img_tuples = []
        for tuple_idx in range(len(img_tuples)):
            if tuple_idx == self._empty_patch_replacement_channel_idx:
                final_img_tuples.append(empty_img_tuples[tuple_idx])
            else:
                final_img_tuples.append(img_tuples[tuple_idx])
        return tuple(final_img_tuples)

    def get_mean_std_for_input(self):
        mean, std = self.get_mean_std()
        return mean["input"], std["input"]

    def _compute_target(self, img_tuples, alpha):
        if self._tar_idx_list is not None and isinstance(self._tar_idx_list, int):
            target = img_tuples[self._tar_idx_list]
        else:
            if self._tar_idx_list is not None:
                assert isinstance(self._tar_idx_list, list) or isinstance(
                    self._tar_idx_list, tuple
                )
                img_tuples = [img_tuples[i] for i in self._tar_idx_list]

            target = np.stack(img_tuples, axis=0)
        return target

    def _compute_input_with_alpha(self, img_tuples, alpha_list):
        # assert self._normalized_input is True, "normalization should happen here"
        if self._input_idx is not None:
            inp = img_tuples[self._input_idx]
        else:
            inp = 0
            for alpha, img in zip(alpha_list, img_tuples):
                inp += img * alpha

            if self._normalized_input is False:
                return inp.astype(np.float32)

        mean, std = self.get_mean_std_for_input()
        mean = mean.squeeze()
        std = std.squeeze()
        if mean.size == 1:
            mean = mean.reshape(
                1,
            )
            std = std.reshape(
                1,
            )

        for i in range(len(mean)):
            assert mean[0] == mean[i]
            assert std[0] == std[i]

        inp = (inp - mean[0]) / std[0]
        return inp.astype(np.float32)

    def _sample_alpha(self):
        alpha_arr = []
        for i in range(self._num_channels):
            alpha_pos = np.random.rand()
            alpha = self._start_alpha_arr[i] + alpha_pos * (
                self._end_alpha_arr[i] - self._start_alpha_arr[i]
            )
            alpha_arr.append(alpha)
        return alpha_arr

    def _compute_input(self, img_tuples):
        alpha = [1 / len(img_tuples) for _ in range(len(img_tuples))]
        if self._start_alpha_arr is not None:
            alpha = self._sample_alpha()

        inp = self._compute_input_with_alpha(img_tuples, alpha)
        if self._input_is_sum:
            inp = len(img_tuples) * inp

        # TODO instead we add channel here
        if len(inp.shape) == 2 or (len(inp.shape) == 3 and self._3Ddata):
            inp = inp[None, ...]

        return inp, alpha

    def _get_index_from_valid_target_logic(self, index):
        if self._validtarget_rand_fract is not None:
            if np.random.rand() < self._validtarget_rand_fract:
                index = self._train_index_switcher.get_valid_target_index()
            else:
                index = self._train_index_switcher.get_invalid_target_index()
        return index

    def _rotate2D(self, img_tuples):
        img_kwargs = {}
        for i, img in enumerate(img_tuples):
            for k in range(len(img)):
                img_kwargs[f"img{i}_{k}"] = img[k]

        keys = list(img_kwargs.keys())
        self._rotation_transform.add_targets({k: "image" for k in keys})
        rot_dic = self._rotation_transform(image=img_tuples[0][0], **img_kwargs)

        rotated_img_tuples = []
        for i, img in enumerate(img_tuples):
            if len(img) == 1:
                rotated_img_tuples.append(rot_dic[f"img{i}_0"][None])
            else:
                rotated_img_tuples.append(
                    np.concatenate(
                        [rot_dic[f"img{i}_{k}"][None] for k in range(len(img))], axis=0
                    )
                )

        return rotated_img_tuples

    def _rotate3D(self, img_tuples):
        img_kwargs = {}
        # random flip in z direction
        flip_z = self._flipz_3D and np.random.rand() < 0.5
        for i, img in enumerate(img_tuples):
            for j in range(self._depth3D):
                for k in range(len(img)):
                    if flip_z:
                        z_idx = self._depth3D - 1 - j
                    else:
                        z_idx = j
                    img_kwargs[f"img{i}_{z_idx}_{k}"] = img[k, j]

        keys = list(img_kwargs.keys())
        self._rotation_transform.add_targets({k: "image" for k in keys})
        rot_dic = self._rotation_transform(image=img_tuples[0][0][0], **img_kwargs)
        rotated_img_tuples = []
        for i, img in enumerate(img_tuples):
            if len(img) == 1:
                rotated_img_tuples.append(
                    np.concatenate(
                        [
                            rot_dic[f"img{i}_{j}_0"][None, None]
                            for j in range(self._depth3D)
                        ],
                        axis=1,
                    )
                )
            else:
                temp_arr = []
                for k in range(len(img)):
                    temp_arr.append(
                        np.concatenate(
                            [
                                rot_dic[f"img{i}_{j}_{k}"][None, None]
                                for j in range(self._depth3D)
                            ],
                            axis=1,
                        )
                    )
                rotated_img_tuples.append(np.concatenate(temp_arr, axis=0))

        return rotated_img_tuples

    def _rotate(self, img_tuples, noise_tuples):

        if self._3Ddata:
            return self._rotate3D(img_tuples, noise_tuples)
        else:
            return self._rotate2D(img_tuples, noise_tuples)

    def _get_img(self, ch_idx: int, patch_idx: int):
        """
        Loads an image.
        Crops the image such that cropped image has content.
        """
        img = self._load_img(ch_idx, patch_idx)
        cropped_img = self._crop_imgs(ch_idx, patch_idx, img)
        return cropped_img

    def get_uncorrelated_img_tuples(self, index):
        """
        Content of channels like actin and nuclei is "correlated" in its
        respective location, this function allows to pick channels' content
        from different patches of the image to make it "uncorrelated".
        """
        img_tuples = []
        for ch_idx in range(len(self._data)):
            if ch_idx == 0:
                # dataset index becomes sample index because all channels have the same
                # length
                img_tuples.append(self._get_img(0, index))
            else:
                # get a random index from corresponding channel
                sample_index = np.random.randint(
                    self.idx_manager.total_grid_count()[0][ch_idx]
                )
                img_tuples.append(self._get_img(ch_idx, sample_index))
        return img_tuples

    def __getitem__(
        self, index: Union[int, tuple[int, int]]
    ) -> tuple[np.ndarray, np.ndarray]:

        # Uncorrelated channels means crops to create the input are taken from different
        # spatial locations of the image.
        if (
            self._uncorrelated_channels
            and np.random.rand() < self._uncorrelated_channel_probab
        ):
            input_tuples = self.get_uncorrelated_img_tuples(index)
        else:
            # 0 is the channel index, because in this case locations are the same for
            # all channels
            # tuple for compatibility with _compute_input. #TODO check
            input_tuples = (self._get_img(0, index),)

        if self._enable_rotation:
            input_tuples = self._rotate(input_tuples)

        # Weight the individual channels, typically alpha is fixed
        inp, alpha = self._compute_input(input_tuples)

        target = self._compute_target(input_tuples, alpha)
        norm_target = self.normalize_target(target)

        return inp, norm_target


class LCMultiChDloaderRef(MultiChDloaderRef):
    def __init__(
        self,
        data_config: MicroSplitDataConfig,
        fpath: str,
        load_data_fn: Callable,
        val_fraction=None,
        test_fraction=None,
    ):
        self._padding_kwargs = (
            data_config.padding_kwargs  # mode=padding_mode, constant_values=constant_value
        )
        self._uncorrelated_channel_probab = data_config.uncorrelated_channel_probab

        super().__init__(
            data_config,
            fpath,
            load_data_fn=load_data_fn,
            val_fraction=val_fraction,
            test_fraction=test_fraction,
        )

        if data_config.overlapping_padding_kwargs is not None:
            assert (
                self._padding_kwargs == data_config.overlapping_padding_kwargs
            ), "During evaluation, overlapping_padding_kwargs should be same as padding_args. \
                It should be so since we just use overlapping_padding_kwargs when it is not None"

        else:
            self._overlapping_padding_kwargs = data_config.padding_kwargs

        self.multiscale_lowres_count = data_config.multiscale_lowres_count
        assert self.multiscale_lowres_count is not None
        self._scaled_data = [self._data]
        self._scaled_noise_data = [self._noise_data]

        assert (
            isinstance(self.multiscale_lowres_count, int)
            and self.multiscale_lowres_count >= 1
        )
        assert isinstance(self._padding_kwargs, dict)
        assert "mode" in self._padding_kwargs

        for _ in range(1, self.multiscale_lowres_count):
            shape = self._scaled_data[-1].shape
            assert len(shape) == 4
            new_shape = (shape[0], shape[1] // 2, shape[2] // 2, shape[3])
            ds_data = resize(
                self._scaled_data[-1].astype(np.float32), new_shape
            ).astype(self._scaled_data[-1].dtype)
            # NOTE: These asserts are important. the resize method expects np.float32. otherwise, one gets weird results.
            assert (
                ds_data.max() / self._scaled_data[-1].max() < 5
            ), "Downsampled image should not have very different values"
            assert (
                ds_data.max() / self._scaled_data[-1].max() > 0.2
            ), "Downsampled image should not have very different values"

            self._scaled_data.append(ds_data)
            # do the same for noise
            if self._noise_data is not None:
                noise_data = resize(self._scaled_noise_data[-1], new_shape)
                self._scaled_noise_data.append(noise_data)

    def reduce_data(
        self, t_list=None, h_start=None, h_end=None, w_start=None, w_end=None
    ):
        assert t_list is not None
        assert h_start is None
        assert h_end is None
        assert w_start is None
        assert w_end is None

        self._data = self._data[t_list].copy()
        self._scaled_data = [
            self._scaled_data[i][t_list].copy() for i in range(len(self._scaled_data))
        ]

        if self._noise_data is not None:
            self._noise_data = self._noise_data[t_list].copy()
            self._scaled_noise_data = [
                self._scaled_noise_data[i][t_list].copy()
                for i in range(len(self._scaled_noise_data))
            ]

        self.N = len(t_list)
        # TODO where tf is self._img_sz defined?
        self.set_img_sz([self._img_sz, self._img_sz], self._grid_sz)
        print(
            f"[{self.__class__.__name__}] Data reduced. New data shape: {self._data.shape}"
        )

    def _init_msg(self):
        msg = super()._init_msg()
        msg += f" Pad:{self._padding_kwargs}"
        if self._uncorrelated_channels:
            msg += f" UncorrChProbab:{self._uncorrelated_channel_probab}"
        return msg

    def _load_scaled_img(
        self, scaled_index, index: Union[int, tuple[int, int]]
    ) -> tuple[np.ndarray, np.ndarray]:
        if isinstance(index, int):
            idx = index
        else:
            idx, _ = index

        # tidx = self.idx_manager.get_t(idx)
        patch_loc_list = self.idx_manager.get_patch_location_from_dataset_idx(idx)
        nidx = patch_loc_list[0]

        imgs = self._scaled_data[scaled_index][nidx]
        imgs = tuple([imgs[None, ..., i] for i in range(imgs.shape[-1])])
        if self._noise_data is not None:
            noisedata = self._scaled_noise_data[scaled_index][nidx]
            noise = tuple([noisedata[None, ..., i] for i in range(noisedata.shape[-1])])
            factor = np.sqrt(2) if self._input_is_sum else 1.0
            imgs = tuple([img + noise[0] * factor for img in imgs])
        return imgs

    def _crop_img(self, img: np.ndarray, patch_start_loc: tuple):
        """
        Here, h_start, w_start could be negative. That simply means we need to pick the content from 0. So,
        the cropped image will be smaller than self._img_sz * self._img_sz
        """
        max_len_vals = list(self.idx_manager.data_shape[1:-1])
        max_len_vals[-2:] = img.shape[-2:]
        return self._crop_img_with_padding(
            img, patch_start_loc, max_len_vals=max_len_vals
        )

    def _get_img(self, index: int):
        """
        Returns the primary patch along with low resolution patches centered on the primary patch.
        """
        # Noise_tuples is populated when there is synthetic noise in training
        # Should have similar type of noise with the noise model
        # Starting with microsplit, dump the noise, use it instead as an augmentation if nessesary
        img_tuples, noise_tuples = self._load_img(index)
        assert self._img_sz is not None
        h, w = img_tuples[0].shape[-2:]
        if self._enable_random_cropping:
            patch_start_loc = self._get_random_hw(h, w)
            if self._3Ddata:
                patch_start_loc = (
                    np.random.choice(img_tuples[0].shape[-3] - self._depth3D),
                ) + patch_start_loc
        else:
            patch_start_loc = self._get_deterministic_loc(index)

        # LC logic is located here, the function crops the image of the highest resolution
        cropped_img_tuples = [
            self._crop_flip_img(img, patch_start_loc, False, False)
            for img in img_tuples
        ]
        cropped_noise_tuples = [
            self._crop_flip_img(noise, patch_start_loc, False, False)
            for noise in noise_tuples
        ]
        patch_start_loc = list(patch_start_loc)
        h_start, w_start = patch_start_loc[-2], patch_start_loc[-1]
        h_center = h_start + self._img_sz // 2
        w_center = w_start + self._img_sz // 2
        allres_versions = {
            i: [cropped_img_tuples[i]] for i in range(len(cropped_img_tuples))
        }
        for scale_idx in range(1, self.multiscale_lowres_count):
            # Returning the image of the lower resolution
            scaled_img_tuples = self._load_scaled_img(scale_idx, index)

            h_center = h_center // 2
            w_center = w_center // 2

            h_start = h_center - self._img_sz // 2
            w_start = w_center - self._img_sz // 2
            patch_start_loc[-2:] = [h_start, w_start]
            scaled_cropped_img_tuples = [
                self._crop_flip_img(img, patch_start_loc, False, False)
                for img in scaled_img_tuples
            ]
            for ch_idx in range(len(img_tuples)):
                allres_versions[ch_idx].append(scaled_cropped_img_tuples[ch_idx])

        output_img_tuples = tuple(
            [
                np.concatenate(allres_versions[ch_idx])
                for ch_idx in range(len(img_tuples))
            ]
        )
        return output_img_tuples, cropped_noise_tuples

    def __getitem__(self, index: Union[int, tuple[int, int]]):
        img_tuples, noise_tuples = self._get_img(index)
        if self._uncorrelated_channels:
            assert (
                self._input_idx is None
            ), "Uncorrelated channels is not implemented when there is a separate input channel."
            if np.random.rand() < self._uncorrelated_channel_probab:
                img_tuples_new = [None] * len(img_tuples)
                img_tuples_new[0] = img_tuples[0]
                for i in range(1, len(img_tuples)):
                    new_index = np.random.randint(len(self))
                    img_tuples_tmp, _ = self._get_img(new_index)
                    img_tuples_new[i] = img_tuples_tmp[i]
                img_tuples = img_tuples_new

        if self._is_train:
            if self._empty_patch_replacement_enabled:
                if np.random.rand() < self._empty_patch_replacement_probab:
                    img_tuples = self.replace_with_empty_patch(img_tuples)

        if self._enable_rotation:
            img_tuples, noise_tuples = self._rotate(img_tuples, noise_tuples)

        # add noise to input, if noise is present combine it with the image
        # factor is for the compute input not to have too much noise because the average of two gaussians
        if len(noise_tuples) > 0:
            factor = np.sqrt(2) if self._input_is_sum else 1.0
            input_tuples = []
            for x in img_tuples:
                x = (
                    x.copy()
                )  # to avoid changing the original image since it is later used for target
                # NOTE: other LC levels already have noise added. So, we just need to add noise to the highest resolution.
                x[0] = x[0] + noise_tuples[0] * factor
                input_tuples.append(x)
        else:
            input_tuples = img_tuples

        # Compute the input by sum / average the channels
        # Alpha is an amount of weight which is applied to the channels when combining them
        # How to sample alpha is still under research
        inp, alpha = self._compute_input(input_tuples)
        target_tuples = [img[:1] for img in img_tuples]
        # add noise to target.
        if len(noise_tuples) >= 1:
            target_tuples = [
                x + noise for x, noise in zip(target_tuples, noise_tuples[1:])
            ]

        target = self._compute_target(target_tuples, alpha)

        norm_target = self.normalize_target(target)

        output = [inp, norm_target]

        if self._return_alpha:
            output.append(alpha)

        if isinstance(index, int):
            return tuple(output)

        _, grid_size = index
        output.append(grid_size)
        return tuple(output)



from collections.abc import Sequence
from pathlib import Path
from typing import TYPE_CHECKING, Any, Literal

import zarr
from numpy.typing import NDArray
from zarr.storage import StorePath

from careamics.config.validators import check_czi_axes_validity
from careamics.file_io import ReadFunc

from ..image_stack import (
    FileImageStack,
    InMemoryImageStack,
    ZarrImageStack,
)
from ..image_stack.czi_image_stack import CziImageStack
from .zarr_utils import collect_arrays, decipher_zarr_uri, is_ome_zarr, is_valid_uri

if TYPE_CHECKING:
    from careamics.file_io.read import ReadFunc


def load_arrays(source: Sequence[NDArray[Any]], axes: str) -> list[InMemoryImageStack]:
    """
    Load image stacks from a sequence of numpy arrays.

    Parameters
    ----------
    source: sequence of numpy.ndarray
        The source arrays of the data.
    axes: str
        The original axes of the data, must be a subset of "STCZYX".

    Returns
    -------
    list[InMemoryImageStack]
    """
    return [InMemoryImageStack.from_array(data=array, axes=axes) for array in source]


# TIFF case
def load_tiffs(source: Sequence[Path], axes: str) -> list[InMemoryImageStack]:
    """
    Load image stacks from a sequence of TIFF files.

    Parameters
    ----------
    source: sequence of Path
        The source files for the data.
    axes: str
        The original axes of the data, must be a subset of "STCZYX".

    Returns
    -------
    list[InMemoryImageStack]
    """
    return [InMemoryImageStack.from_tiff(path=path, axes=axes) for path in source]


# TODO: better name
# iter Tiff
def load_iter_tiff(source: Sequence[Path], axes: str) -> list[FileImageStack]:
    # TODO: better docs
    """
    Load image stacks from a sequence of TIFF files.

    Parameters
    ----------
    source: sequence of Path
        The source files for the data.
    axes: str
        The original axes of the data, must be a subset of "STCZYX".

    Returns
    -------
    list[FileImageStack]
    """
    return [FileImageStack.from_tiff(path=path, axes=axes) for path in source]


# Custom file type case (loaded into memory)
def load_custom_file(
    source: Sequence[Path],
    axes: str,
    *,
    read_func: ReadFunc,
    read_kwargs: dict[str, Any],
) -> list[InMemoryImageStack]:
    """
    Load image stacks from a sequence of files of a custom type.

    Parameters
    ----------
    source: sequence of Path
        The source files for the data.
    axes: str
        The original axes of the data, must be a subset of "STCZYX".
    read_func : ReadFunc
        A function to read the custom file type, see the `ReadFunc` protocol.
    read_kwargs : dict of {str: Any}
        Kwargs that will be passed to the custom `read_func`.

    Returns
    -------
    list[InMemoryImageStack]
    """
    # TODO: lazy loading custom files
    return [
        InMemoryImageStack.from_custom_file_type(
            path=path,
            axes=axes,
            read_func=read_func,
            **read_kwargs,
        )
        for path in source
    ]


def load_zarrs(
    source: Sequence[str | Path | StorePath],
    axes: str,
) -> list[ZarrImageStack]:
    """Create a list of ZarrImageStack from a sequence of zarr file paths or URIs.

    File paths must point to a zarr store (ending with .zarr) and URIs must be in the
    format "file://path/to/zarr_store.zarr/group/path/array_name".

    If the zarr file is an OME-Zarr, the specified multiscale level will be used. Note
    that OME-Zarrs are only supported when providing a path to the zarr store, not when
    using a file URI. One can, however, provide a file URI to the specific resolution
    array within the OME-Zarr.

    Parameters
    ----------
    source : sequence of str or Path
        The source zarr file paths or URIs.
    axes : str
        The original axes of the data, must be a subset of "STCZYX".

    Returns
    -------
    list of ZarrImageStack
        A list of ZarrImageStack created from the sources.
    """

    image_stacks: list[ZarrImageStack] = []

    for data_source in source:
        data_str = str(data_source)

        # either a path to a zarr file or a uri "file://path/to/zarr/array_path"
        if data_str.endswith(".zarr"):
            zarr_group = zarr.open_group(data_str, mode="r")

            # test if ome-zarr (minimum assumption of multiscales)
            if is_ome_zarr(zarr_group):
                # TODO placeholder for handling OME-Zarr
                # - Need to potentially select multiscale level
                # - Extract axes and compare with provided ones
                raise NotImplementedError(
                    "OME-Zarr support is not yet implemented when providing a "
                    "path to the zarr store. Please provide a file URI to the "
                    "specific array within the OME-Zarr."
                )
            else:
                # collect all arrays
                array_paths = collect_arrays(zarr_group)

                # sort names
                array_paths.sort()

            # instantiate image stacks
            for array_path in array_paths:
                image_stacks.append(
                    ZarrImageStack(group=zarr_group, data_path=array_path, axes=axes)
                )

        elif is_valid_uri(data_str):
            # decipher the uri and open the group
            store_path, parent_path, name = decipher_zarr_uri(data_str)

            zarr_group = zarr.open_group(store_path, path=parent_path, mode="r")
            content = zarr_group[name]

            # assert if group or array
            if isinstance(content, zarr.Group):
                array_paths = collect_arrays(content)

                # sort the names
                array_paths.sort()

                for array_path in array_paths:
                    image_stacks.append(
                        ZarrImageStack(group=content, data_path=array_path, axes=axes)
                    )
            else:
                if not isinstance(content, zarr.Array):
                    raise TypeError(
                        f"Content at '{data_str}' is neither a zarr.Group nor "
                        f"a zarr.Array."
                    )

                # create image stack from a single array
                image_stacks.append(
                    ZarrImageStack(
                        group=zarr_group,
                        data_path=name,
                        axes=axes,
                    )
                )

        else:
            raise ValueError(
                f"Source '{data_source}' is neither a zarr file nor a file URI."
            )

    return image_stacks


def load_czis(
    source: Sequence[Path],
    axes: str,
) -> list[CziImageStack]:
    """
    Load CZI image stacks from a sequence of CZI files paths.

    If the CZI files contain multiple scenes, one image stack will be created for
    each scene.

    Axes should be in the format "SC(Z/T)YX", where Z or T are optional, and S and C
    can be singleton dimensions, but must be provided.

    Parameters
    ----------
    source: sequence of Path
        The source files for the data.
    axes: str
        Specifies which axes of the data to use and how.
        If this string ends with `"ZYX"` or `"TYX"`, the data will consist of 3-D
        patches, using `Z` or `T` as third dimension, respectively.
        If the string does not end with "ZYX", the data will consist of 2-D patches.

    Returns
    -------
    list[CziImageStack]

    Raises
    ------
    ValueError
        If the provided axes are not valid.
    """
    if check_czi_axes_validity(axes) is False:
        raise ValueError(
            f"Provided axes '{axes}' are not valid. Axes must be in the `SC(Z/T)YX` "
            f"format, where Z or T are optional, and S and C can be singleton "
            f"dimensions, but must be provided."
        )

    depth_axis: Literal["none", "Z", "T"] = "none"
    if axes.endswith("TYX"):
        depth_axis = "T"
    elif axes.endswith("ZYX"):
        depth_axis = "Z"

    image_stacks: list[CziImageStack] = []
    for path in source:
        scene_rectangles = CziImageStack.get_bounding_rectangles(path)
        image_stacks.extend(
            CziImageStack(path, scene=scene, depth_axis=depth_axis)
            for scene in scene_rectangles.keys()
        )
    return image_stacks



from pathlib import Path

import numpy as np
import tifffile

from careamics.dataset_ng.image_stack import FileImageStack
from careamics.dataset_ng.patch_extractor.limit_file_extractor import (
    LimitFilesPatchExtractor,
)
from careamics.dataset_ng.patching_strategies import RandomPatchingStrategy


def test_files_limited(tmp_path: Path):
    """Test that `LimitFilesPatchExtractor` never has more than two files loaded."""

    rng = np.random.default_rng(42)

    # set up test data
    data_shapes = [(64, 48), (55, 54), (71, 65), (32, 32)]
    input_data = [np.arange(np.prod(shape)).reshape(shape) for shape in data_shapes]
    paths = [tmp_path / f"{i}.tiff" for i in range(len(data_shapes))]
    for path, data in zip(paths, input_data, strict=True):
        tifffile.imwrite(path, data, metadata={"axes": "YX"})

    # set up patch extractor
    image_stacks = [FileImageStack.from_tiff(path, axes="YX") for path in paths]
    patch_extractor = LimitFilesPatchExtractor(image_stacks)

    # using random patching
    patching = RandomPatchingStrategy(
        [image_stack.data_shape for image_stack in image_stacks],
        patch_size=(16, 16),
        seed=42,
    )
    patch_specs = [patching.get_patch_spec(i) for i in range(patching.n_patches)]
    # shuffle the indices because adjacent indices still belong to the same image
    indices = np.arange(len(patch_specs))
    rng.shuffle(indices)

    for i in indices:
        patch_extractor.extract_patch(**patch_specs[i])
        is_loaded = [
            image_stack.is_loaded for image_stack in patch_extractor.image_stacks
        ]
        # 1 or less files should be loaded at any time
        assert np.count_nonzero(is_loaded) <= 1



from pathlib import Path

import numpy as np
import pytest
import tifffile

from careamics.dataset_ng.image_stack import FileImageStack
from careamics.dataset_ng.image_stack.image_utils import channel_slice


def test_extract_patch(tmp_path: Path):
    data_shape = (64, 47)
    axes = "YX"
    data = np.arange(np.prod(data_shape)).reshape(data_shape)
    path = tmp_path / "image.tiff"
    tifffile.imwrite(path, data, metadata={"axes": axes})

    image_stack = FileImageStack.from_tiff(path, axes)

    # extract patch should raise an error if the image stack is not loaded
    with pytest.raises(ValueError):
        image_stack.extract_patch(
            sample_idx=0, channels=None, coords=(4, 8), patch_size=(16, 16)
        )

    # call load & extract patch
    image_stack.load()
    patch = image_stack.extract_patch(
        sample_idx=0, channels=None, coords=(4, 8), patch_size=(16, 16)
    )
    # confirm as expected against reference data
    np.testing.assert_array_equal(patch, data[np.newaxis, 4:20, 8:24])


@pytest.mark.parametrize(
    "shape, axes, channels",
    [
        ((2, 1, 64, 64), "SCYX", None),
        ((2, 1, 64, 64), "SCYX", [0]),
        ((2, 3, 64, 64), "SCYX", None),
        ((2, 3, 64, 64), "SCYX", [0, 2]),
        ((2, 3, 64, 64), "SCYX", [1]),
    ],
)
def test_extract_channels(tmp_path, shape, axes, channels):
    data = np.arange(np.prod(shape)).reshape(shape)
    path = tmp_path / "image.tiff"
    tifffile.imwrite(path, data, metadata={"axes": axes})

    image_stack = FileImageStack.from_tiff(path, axes)

    # call load & extract patch
    image_stack.load()
    patch = image_stack.extract_patch(
        sample_idx=1,
        channels=channels,
        coords=(10, 10),
        patch_size=(32, 32),
    )
    assert len(patch.shape) == 3  # no Z
    assert patch.shape[0] == len(channels) if channels is not None else data.shape[1]

    expected_patch = data[
        1,
        channel_slice(channels),
        10 : 10 + 32,
        10 : 10 + 32,
    ]
    np.testing.assert_array_equal(patch, expected_patch)


@pytest.mark.parametrize(
    "shape, axes, channels",
    [
        ((2, 3, 64, 64), "SCYX", [0, 4]),
        ((2, 3, 64, 64), "SCYX", [3]),
    ],
)
def test_extract_channel_error(
    tmp_path: Path,
    shape: tuple[int, ...],
    axes: str,
    channels: int,
):
    data = np.arange(np.prod(shape)).reshape(shape)
    path = tmp_path / "image.tiff"
    tifffile.imwrite(path, data, metadata={"axes": axes})

    image_stack = FileImageStack.from_tiff(path, axes)

    # call load & extract patch
    image_stack.load()

    expected_msg = (
        f"Channel index {channels[-1]} is out of bounds for data with "
        f"{shape[1]} channels. Check the provided `channels` "
        f"parameter in the configuration for erroneous channel "
        f"indices."
    )

    with pytest.raises(ValueError, match=expected_msg):
        image_stack.extract_patch(
            sample_idx=0,
            channels=channels,
            coords=(0, 0),
            patch_size=(16, 16),
        )


def test_load_and_close(tmp_path: Path):
    data_shape = (64, 47)
    axes = "YX"
    data = np.arange(np.prod(data_shape)).reshape(data_shape)
    path = tmp_path / "image.tiff"
    tifffile.imwrite(path, data, metadata={"axes": axes})

    image_stack = FileImageStack.from_tiff(path, axes)
    assert not image_stack.is_loaded

    image_stack.load()
    assert image_stack.is_loaded

    image_stack.close()
    assert not image_stack.is_loaded



"""Convenience functions to create coherent configurations for CAREamics."""

__all__ = [
    "create_advanced_n2v_config",
    "create_n2v_config",
    "create_ng_data_configuration",
    "create_structn2v_config",
]

from .data_factory import create_ng_data_configuration
from .n2v_factory import (
    create_advanced_n2v_config,
    create_n2v_config,
    create_structn2v_config,
)



"""Pydantic models for normalization strategies."""

from typing import Annotated, Literal, Self, Union

from pydantic import BaseModel, ConfigDict, Discriminator, Field, model_validator


class MeanStdConfig(BaseModel):
    """
    Mean and standard deviation normalization configuration.

    Holds mean and standard deviation statistics for input and target, used to
    normalize data. If not provided, statistics can be computed later.

    Attributes
    ----------
    name : Literal["mean_std"]
        Identifier for the mean-std normalization scheme.
    input_means : list[float] | None
        Means of input channels/features, or None for automatic computation.
    input_stds : list[float] | None
        Standard deviations of input channels/features,
        or None for automatic computation.
    target_means : list[float] | None
        Means of target channels/features, or None for automatic computation.
    target_stds : list[float] | None
        Standard deviations of target channels/features,
        or None for automatic computation.
    """

    model_config = ConfigDict(validate_assignment=True)

    name: Literal["mean_std"] = "mean_std"
    input_means: list[float] | None = Field(default=None, min_length=0, max_length=32)
    input_stds: list[float] | None = Field(default=None, min_length=0, max_length=32)
    target_means: list[float] | None = Field(default=None, min_length=0, max_length=32)
    target_stds: list[float] | None = Field(default=None, min_length=0, max_length=32)

    @model_validator(mode="after")
    def validate_means_stds(self: Self) -> Self:
        """Validate that means and stds are provided in pairs or set to None.

        Returns
        -------
        Self
            The validated model instance.

        Raises
        ------
        ValueError
            If only one of means or stds is provided for input or target,
            or if each pair has mismatched lengths.
        """
        if (self.input_means is None) != (self.input_stds is None):
            raise ValueError(
                "input_means and input_stds must be both provided or both None."
            )
        if self.input_means is not None and self.input_stds is not None:
            if len(self.input_means) != len(self.input_stds):
                raise ValueError("input_means and input_stds must have same length.")

        if (self.target_means is None) != (self.target_stds is None):
            raise ValueError(
                "target_means and target_stds must be both provided or both None."
            )
        if self.target_means is not None and self.target_stds is not None:
            if len(self.target_means) != len(self.target_stds):
                raise ValueError("target_means and target_stds must have same length.")

        return self

    def needs_computation(self) -> bool:
        """
        Check if statistics need to be computed.

        Returns
        -------
        bool
            True if input statistics are missing, False otherwise.
        """
        return self.input_means is None or self.input_stds is None

    def set_input_stats(self, means: list[float], stds: list[float]) -> None:
        """
        Set input means and stds together to avoid validation errors.

        Parameters
        ----------
        means : list[float]
            Mean values per channel.
        stds : list[float]
            Standard deviation values per channel.
        """
        self.__dict__["input_means"] = means
        self.__dict__["input_stds"] = stds
        self.__class__.model_validate(self)

    def set_target_stats(self, means: list[float], stds: list[float]) -> None:
        """
        Set target means and stds together to avoid validation errors.

        Parameters
        ----------
        means : list[float]
            Mean values per channel.
        stds : list[float]
            Standard deviation values per channel.
        """
        self.__dict__["target_means"] = means
        self.__dict__["target_stds"] = stds
        self.__class__.model_validate(self)


class QuantileConfig(BaseModel):
    """
    Quantile normalization configuration.

    Normalizes data using quantile-based range scaling. Quantile levels can be
    specified as a single value (applied to all channels) or a list (one per channel).

    Attributes
    ----------
    name : Literal["quantile"]
        Identifier for quantile normalization.
    lower_quantile : float | list[float]
        Lower quantile level(s). Single float applies to all channels,
        or list for per-channel levels. Values must be in [0, 1).
    upper_quantile : float | list[float]
        Upper quantile level(s). Single float applies to all channels,
        or list for per-channel levels. Values must be in (0, 1].
    input_lower_quantile_values : list[float] | None
        Computed lower quantile values for each input channel.
    input_upper_quantile_values : list[float] | None
        Computed upper quantile values for each input channel.
    target_lower_quantile_values : list[float] | None
        Computed lower quantile values for each target channel.
    target_upper_quantile_values : list[float] | None
        Computed upper quantile values for each target channel.
    """

    model_config = ConfigDict(validate_assignment=True)

    name: Literal["quantile"] = "quantile"

    lower_quantile: float | list[float] = Field(default=0.01)
    upper_quantile: float | list[float] = Field(default=0.99)

    input_lower_quantile_values: list[float] | None = Field(default=None)
    input_upper_quantile_values: list[float] | None = Field(default=None)
    target_lower_quantile_values: list[float] | None = Field(default=None)
    target_upper_quantile_values: list[float] | None = Field(default=None)

    @model_validator(mode="after")
    def validate_quantile_levels(self: Self) -> Self:
        """Validate quantile levels are in valid range and properly ordered.

        Returns
        -------
        Self
            The validated model instance.
        """
        lower = (
            self.lower_quantile
            if isinstance(self.lower_quantile, list)
            else [self.lower_quantile]
        )
        upper = (
            self.upper_quantile
            if isinstance(self.upper_quantile, list)
            else [self.upper_quantile]
        )

        for lq in lower:
            if not (0.0 <= lq < 1.0):
                raise ValueError(f"lower_quantile values must be in [0, 1), got {lq}")
        for uq in upper:
            if not (0.0 < uq <= 1.0):
                raise ValueError(f"upper_quantile values must be in (0, 1], got {uq}")

        if len(lower) != len(upper):
            raise ValueError(
                f"lower_quantile and upper_quantile lists must have same length, "
                f"got {len(lower)} and {len(upper)}"
            )
        for i, (lq, uq) in enumerate(zip(lower, upper, strict=True)):
            if lq >= uq:
                raise ValueError(
                    f"lower_quantile[{i}] ({lq}) must be less than "
                    f"upper_quantile[{i}] ({uq})"
                )
        return self

    @model_validator(mode="after")
    def validate_quantile_values(self: Self) -> Self:
        """Validate that computed quantile value lists are provided in pairs.

        Returns
        -------
        Self
            The validated model instance.
        """
        if (self.input_lower_quantile_values is None) != (
            self.input_upper_quantile_values is None
        ):
            raise ValueError(
                "input_lower_quantile_values and input_upper_quantile_values "
                "must be both provided or both None."
            )
        if (
            self.input_lower_quantile_values is not None
            and self.input_upper_quantile_values is not None
        ):
            if len(self.input_lower_quantile_values) != len(
                self.input_upper_quantile_values
            ):
                raise ValueError(
                    "input_lower_quantile_values and input_upper_quantile_values "
                    "must have same length."
                )
            for i, (lower, upper) in enumerate(
                zip(
                    self.input_lower_quantile_values,
                    self.input_upper_quantile_values,
                    strict=True,
                )
            ):
                if lower >= upper:
                    raise ValueError(
                        f"input_lower_quantile_values[{i}] ({lower}) must be less than "
                        f"input_upper_quantile_values[{i}] ({upper})"
                    )

        if (self.target_lower_quantile_values is None) != (
            self.target_upper_quantile_values is None
        ):
            raise ValueError(
                "target_lower_quantile_values and target_upper_quantile_values "
                "must be both provided or both None."
            )
        if (
            self.target_lower_quantile_values is not None
            and self.target_upper_quantile_values is not None
        ):
            if len(self.target_lower_quantile_values) != len(
                self.target_upper_quantile_values
            ):
                raise ValueError(
                    "target_lower_quantile_values and target_upper_quantile_values "
                    "must have same length."
                )
            for i, (lower, upper) in enumerate(
                zip(
                    self.target_lower_quantile_values,
                    self.target_upper_quantile_values,
                    strict=True,
                )
            ):
                if lower >= upper:
                    raise ValueError(
                        f"target_lower_quantile_values[{i}] ({lower}) must be less "
                        f"than target_upper_quantile_values[{i}] ({upper})"
                    )
        return self

    def get_lower_quantiles_for_channels(self, n_channels: int) -> list[float]:
        """Get lower quantile levels expanded to n_channels.

        Parameters
        ----------
        n_channels : int
            Number of channels in the data.

        Returns
        -------
        list[float]
            Lower quantile levels for each channel.
        """
        if isinstance(self.lower_quantile, list):
            if len(self.lower_quantile) != n_channels:
                raise ValueError(
                    f"lower_quantile has {len(self.lower_quantile)} values but "
                    f"data has {n_channels} channels"
                )
            return self.lower_quantile
        return [self.lower_quantile] * n_channels

    def get_upper_quantiles_for_channels(self, n_channels: int) -> list[float]:
        """Get upper quantile levels expanded to n_channels.

        Parameters
        ----------
        n_channels : int
            Number of channels in the data.

        Returns
        -------
        list[float]
            Upper quantile levels for each channel.
        """
        if isinstance(self.upper_quantile, list):
            if len(self.upper_quantile) != n_channels:
                raise ValueError(
                    f"upper_quantile has {len(self.upper_quantile)} values but "
                    f"data has {n_channels} channels"
                )
            return self.upper_quantile
        return [self.upper_quantile] * n_channels

    def needs_computation(self) -> bool:
        """Check if quantile values need to be computed.

        Returns
        -------
        bool
            True if quantile values need to be computed.
        """
        return (
            self.input_lower_quantile_values is None
            or self.input_upper_quantile_values is None
        )

    def set_input_quantile_values(self, lower: list[float], upper: list[float]) -> None:
        """
        Set input quantile values together to avoid validation errors.

        Parameters
        ----------
        lower : list[float]
            Lower quantile values per channel.
        upper : list[float]
            Upper quantile values per channel.
        """
        self.__dict__["input_lower_quantile_values"] = lower
        self.__dict__["input_upper_quantile_values"] = upper
        self.__class__.model_validate(self)

    def set_target_quantile_values(
        self, lower: list[float], upper: list[float]
    ) -> None:
        """
        Set target quantile values together to avoid validation errors.

        Parameters
        ----------
        lower : list[float]
            Lower quantile values per channel.
        upper : list[float]
            Upper quantile values per channel.
        """
        self.__dict__["target_lower_quantile_values"] = lower
        self.__dict__["target_upper_quantile_values"] = upper
        self.__class__.model_validate(self)


class MinMaxConfig(BaseModel):
    """
    Min-max normalization configuration.

    Stores minimum and maximum statistics for scaling data into a desired range.
    If not provided, statistics can be computed from the data.

    Attributes
    ----------
    name : Literal["minmax"]
        Identifier for min-max normalization.
    input_mins : list[float] | None
        Minimum values for input channels/features.
    input_maxes : list[float] | None
        Maximum values for input channels/features.
    target_mins : list[float] | None
        Minimum values for target channels/features.
    target_maxes : list[float] | None
        Maximum values for target channels/features.
    """

    model_config = ConfigDict(validate_assignment=True)

    name: Literal["minmax"] = "minmax"
    input_mins: list[float] | None = Field(default=None)
    input_maxes: list[float] | None = Field(default=None)
    target_mins: list[float] | None = Field(default=None)
    target_maxes: list[float] | None = Field(default=None)

    @model_validator(mode="after")
    def validate_mins_maxes(self: Self) -> Self:
        """Validate that mins and maxes are provided in pairs or both None.

        Returns
        -------
        Self
            The validated model instance.
        """
        if (self.input_mins is None) != (self.input_maxes is None):
            raise ValueError(
                "input_mins and input_maxes must be both provided or both None."
            )
        if self.input_mins is not None and self.input_maxes is not None:
            if len(self.input_mins) != len(self.input_maxes):
                raise ValueError("input_mins and input_maxes must have same length.")

        if (self.target_mins is None) != (self.target_maxes is None):
            raise ValueError(
                "target_mins and target_maxes must be both provided or both None."
            )
        if self.target_mins is not None and self.target_maxes is not None:
            if len(self.target_mins) != len(self.target_maxes):
                raise ValueError("target_mins and target_maxes must have same length.")

        return self

    def needs_computation(self) -> bool:
        """
        Check if min/max values need to be computed.

        Returns
        -------
        bool
            True if input statistics are missing, False otherwise.
        """
        return self.input_mins is None or self.input_maxes is None

    def set_input_range(self, mins: list[float], maxes: list[float]) -> None:
        """
        Set input mins and maxes together to avoid validation errors.

        Parameters
        ----------
        mins : list[float]
            Minimum values per channel.
        maxes : list[float]
            Maximum values per channel.
        """
        self.__dict__["input_mins"] = mins
        self.__dict__["input_maxes"] = maxes
        self.__class__.model_validate(self)

    def set_target_range(self, mins: list[float], maxes: list[float]) -> None:
        """
        Set target mins and maxes together to avoid validation errors.

        Parameters
        ----------
        mins : list[float]
            Minimum values per channel.
        maxes : list[float]
            Maximum values per channel.
        """
        self.__dict__["target_mins"] = mins
        self.__dict__["target_maxes"] = maxes
        self.__class__.model_validate(self)


class NoNormConfig(BaseModel):
    """
    No normalization configuration.

    Indicates that no normalization should be applied.

    Attributes
    ----------
    name : Literal["none"]
        Identifier for no normalization scheme.
    """

    name: Literal["none"] = "none"

    def needs_computation(self) -> bool:
        """Check if statistics need to be computed.

        Returns
        -------
        bool
            Always False, as no statistics are required.
        """
        return False


NormalizationConfig = Annotated[
    Union[MeanStdConfig, NoNormConfig, QuantileConfig, MinMaxConfig],
    Discriminator("name"),
]



"""Generic patching Pydantic model."""

from collections.abc import Sequence

from pydantic import BaseModel, ConfigDict, Field, field_validator

from careamics.config.validators import patch_size_ge_than_8_power_of_2


class _PatchedConfig(BaseModel):
    """Generic patching Pydantic model.

    This model is only used for inheritance and validation purposes.
    """

    model_config = ConfigDict(
        extra="ignore",  # default behaviour, make it explicit
    )

    name: str
    """The name of the patching strategy."""

    patch_size: Sequence[int] = Field(..., min_length=2, max_length=3)
    """The size of the patch in each spatial dimensions, each patch size must be a power
    of 2 and larger than 8."""

    @field_validator("patch_size")
    @classmethod
    def all_elements_power_of_2_minimum_8(
        cls, patch_list: Sequence[int]
    ) -> Sequence[int]:
        """
        Validate patch size.

        Patch size must be powers of 2 and minimum 8.

        Parameters
        ----------
        patch_list : Sequence of int
            Patch size.

        Returns
        -------
        Sequence of int
            Validated patch size.

        Raises
        ------
        ValueError
            If the patch size is smaller than 8.
        ValueError
            If the patch size is not a power of 2.
        """
        patch_size_ge_than_8_power_of_2(patch_list)

        return patch_list



from collections.abc import Sequence
from pathlib import Path
from typing import Any, Literal, Self, Union

import numpy as np
from numpy.typing import DTypeLike, NDArray

from careamics.dataset.dataset_utils import reshape_array
from careamics.file_io.read import ReadFunc, read_tiff

from .image_utils.image_stack_utils import channel_slice, pad_patch


class InMemoryImageStack:
    """
    A class for extracting patches from an image stack that has been loaded into memory.
    """

    def __init__(self, source: Union[Path, Literal["array"]], data: NDArray):
        self.source: Union[str, Path, Literal["array"]] = source
        # data expected to be in SC(Z)YX shape, reason to use from_array constructor
        self._data: NDArray = data
        self.data_shape: Sequence[int] = self._data.shape
        self.data_dtype: DTypeLike = self._data.dtype

    def extract_patch(
        self,
        sample_idx: int,
        channels: Sequence[int] | None,  # `channels = None` to select all channels
        coords: Sequence[int],
        patch_size: Sequence[int],
    ) -> NDArray:
        if (coord_dims := len(coords)) != (patch_dims := len(patch_size)):
            raise ValueError(
                "Patch coordinates and patch size must have the same dimensions but "
                f"found {coord_dims} ({coords}) and {patch_dims} ({patch_size})."
            )

        # check that channels are within bounds
        if channels is not None:
            max_channel = self.data_shape[1] - 1  # channel is second dimension
            for ch in channels:
                if ch > max_channel:
                    raise ValueError(
                        f"Channel index {ch} is out of bounds for data with "
                        f"{self.data_shape[1]} channels. Check the provided `channels` "
                        f"parameter in the configuration for erroneous channel "
                        f"indices."
                    )

        # TODO: test for 2D or 3D?

        patch_data = self._data[
            (
                sample_idx,  # type: ignore
                # use channel slice so that channel dimension is kept
                channel_slice(channels),  # type: ignore
                *[
                    slice(
                        np.clip(c, 0, self.data_shape[2 + i]),
                        np.clip(c + ps, 0, self.data_shape[2 + i]),
                    )
                    for i, (c, ps) in enumerate(zip(coords, patch_size, strict=False))
                ],  # type: ignore
            )  # type: ignore
        ]
        patch = pad_patch(coords, patch_size, self.data_shape, patch_data)

        return patch

    @classmethod
    def from_array(cls, data: NDArray, axes: str) -> Self:
        data = reshape_array(data, axes)
        return cls(source="array", data=data)

    @classmethod
    def from_tiff(cls, path: Path, axes: str) -> Self:
        data = read_tiff(path)
        data = reshape_array(data, axes)
        return cls(source=path, data=data)

    @classmethod
    def from_custom_file_type(
        cls, path: Path, axes: str, read_func: ReadFunc, **read_kwargs: Any
    ) -> Self:
        data = read_func(path, **read_kwargs)
        data = reshape_array(data, axes)
        return cls(source=path, data=data)



"""
A module for utility functions that adapts the new dataset outputs to work with previous
code until it is updated.
"""

from collections.abc import Sequence
from typing import cast

import numpy as np
from numpy.typing import NDArray

from careamics.config.data.tile_information import TileInformation
from careamics.utils.deprecation import deprecated

from .dataset import ImageRegionData
from .patching_strategies import TileSpecs


@deprecated(
    "use careamics.lightning.dataset_ng.prediction functions instead for stitching and"
    " combining predictions."
)
def imageregions_to_tileinfos(
    image_regions: Sequence[ImageRegionData],
) -> list[tuple[NDArray, list[TileInformation]]]:
    """
    Converts a series of `TileSpecs` dictionaries to `TileInformation` pydantic class.

    Parameters
    ----------
    image_regions : sequence of ImageRegionData
        A list of ImageRegionData, it must have an instance of `TileSpecs` as it's
        `region_data` field.

    Returns
    -------
    list of TileInformation
        The converted tile information.
    """

    tile_infos: list[TileInformation] = []

    data = [image_region.data for image_region in image_regions]
    tile_specs = [image_region.region_spec for image_region in image_regions]

    data_indices: NDArray[np.int_] = np.array(
        [tile_spec["data_idx"] for tile_spec in tile_specs], dtype=int
    )
    unique_data_indices = np.unique(data_indices)
    # data_idx denotes which image stack a patch belongs to
    # separate TileSpecs by image_stack
    for data_idx in unique_data_indices:
        # collect all ImageRegions
        data_image_regions: list[ImageRegionData] = [
            image_region
            for image_region in image_regions
            if image_region.region_spec["data_idx"] == data_idx
        ]

        # --- find last indices
        # make sure tiles belonging to the same sample are together
        data_image_regions.sort(
            key=lambda image_region: image_region.region_spec["sample_idx"]
        )
        sample_indices = np.array(
            [
                image_region.region_spec["sample_idx"]
                for image_region in data_image_regions
            ]
        )
        # reverse array so indices returned are at far edge
        _, unique_indices = np.unique(sample_indices[::-1], return_index=True)
        # un reverse indices
        last_indices = len(sample_indices) - 1 - unique_indices

        # convert each ImageRegionData to tile_info
        for i, image_region in enumerate(data_image_regions):
            last_tile = i in last_indices
            tile_info = _imageregion_to_tileinfo(image_region, last_tile)
            tile_infos.append(tile_info)

    return [
        (data, [tile_info]) for data, tile_info in zip(data, tile_infos, strict=False)
    ]


def _imageregion_to_tileinfo(
    image_region: ImageRegionData, last_tile: bool
) -> TileInformation:
    """
    Convert a single `ImageRegionData` instance to a `TileInformation` instance. Whether
    it is the last tile in a sequence needs to be supplied.

    Parameters
    ----------
    image_region : ImageRegionData
        An instance of `ImageRegionData`, it must have an instance of `TileSpecs` as
        it's `region_data` field.
    last_tile : bool
        Whether a tile is the last tile in a sequence, for stitching.

    Returns
    -------
    TileInformation
        A tile information object.

    Raises
    ------
    KeyError
        If `image_region.region_spec` does not contain the keys: {'crop_coords',
        'crop_size', 'stitch_coords'}.
    """
    patch_spec = image_region.region_spec
    data_shape = image_region.data_shape

    # TODO: In python 3.11 and greater, NamedTuples can inherit from Generic
    #   so we could do image_region: ImageRegionData[TileSpecs]
    #   and not have to do this check here + cast
    # make sure image_region.region_spec is TileSpec
    if (
        ("crop_coords" not in patch_spec)
        or ("crop_size" not in patch_spec)
        or ("stitch_coords" not in patch_spec)
    ):
        raise KeyError(
            "Could not find all keys: {'crop_coords', 'crop_size', 'stitch_coords'} in "
            "`image_region.region_spec`."
        )
    tile_spec = cast(TileSpecs, patch_spec)  # ugly cast for mypy
    return _tilespec_to_tileinfo(tile_spec, data_shape, last_tile)


def _tilespec_to_tileinfo(
    tile_spec: TileSpecs, data_shape: Sequence[int], last_tile: bool
) -> TileInformation:
    """
    Convert a single `TileSpec` to a `TileInformation`. Whether it is the last tile
    needs to be supplied.

    Parameters
    ----------
    tile_spec : TileSpecs
        A tile spec dictionary.
    data_shape : sequence of int
        The original shape of the data the tile came from, labeling the dimensions of
        axes SC(Z)YX.
    last_tile : bool
        Whether a tile is the last tile in a sequence, for stitching.

    Returns
    -------
    TileInformation
        A tile information object.
    """
    overlap_crop_coords = tuple(
        (
            tile_spec["crop_coords"][i],
            tile_spec["crop_coords"][i] + tile_spec["crop_size"][i],
        )
        for i in range(len(tile_spec["crop_coords"]))
    )
    stitch_coords = tuple(
        (
            tile_spec["stitch_coords"][i],
            tile_spec["stitch_coords"][i] + tile_spec["crop_size"][i],
        )
        for i in range(len(tile_spec["crop_coords"]))
    )
    return TileInformation(
        array_shape=tuple(data_shape[1:]),  # remove sample dimension
        last_tile=last_tile,
        overlap_crop_coords=overlap_crop_coords,
        stitch_coords=stitch_coords,
        sample_id=tile_spec["sample_idx"],
    )



"""Axes validation utilities."""

_AXES = "STCZYX"


def check_axes_validity(axes: str) -> None:
    """
    Sanity check on axes.

    The constraints on the axes are the following:
    - must be a combination of 'STCZYX'
    - must not contain duplicates
    - must contain at least 2 contiguous axes: X and Y
    - must contain at most 4 axes

    Axes do not need to be in the order 'STCZYX', as this depends on the user data.

    Parameters
    ----------
    axes : str
        Axes to validate.
    """
    _axes = axes.upper()

    # Minimum is 2 (XY) and maximum is 4 (TZYX)
    if len(_axes) < 2 or len(_axes) > 6:
        raise ValueError(
            f"Invalid axes {axes}. Must contain at least 2 and at most 6 axes."
        )

    if "YX" not in _axes and "XY" not in _axes:
        raise ValueError(
            f"Invalid axes {axes}. Must contain at least X and Y axes consecutively."
        )

    # all characters must be in REF_AXES = 'STCZYX'
    if not all(s in _AXES for s in _axes):
        raise ValueError(f"Invalid axes {axes}. Must be a combination of {_AXES}.")

    # check for repeating characters
    for i, s in enumerate(_axes):
        if i != _axes.rfind(s):
            raise ValueError(
                f"Invalid axes {axes}. Cannot contain duplicate axes"
                f" (got multiple {axes[i]})."
            )


def check_czi_axes_validity(axes: str) -> bool:
    """
    Check if the provided axes string is valid for CZI files.

    CZI axes is always in the "SC(Z/T)YX" format, where Z or T are optional, and S and C
    can be singleton dimensions, but must be provided.

    Parameters
    ----------
    axes : str
        The axes string to validate.

    Returns
    -------
    bool
        True if the axes string is valid, False otherwise.
    """
    valid_axes = {"S", "C", "Z", "T", "Y", "X"}
    axes_set = set(axes)

    # check for invalid characters
    if not axes_set.issubset(valid_axes):
        return False

    # check for mandatory axes
    if not ({"S", "C", "Y", "X"}.issubset(axes_set)):
        return False

    # check for mutually exclusive axes
    if "Z" in axes_set and "T" in axes_set:
        return False

    # check for correct order
    order = "SCZYX" if "Z" in axes else "SCTYX"
    last_index = -1
    for axis in axes:
        current_index = order.find(axis)
        if current_index < last_index:
            return False
        last_index = current_index

    return True



import numpy as np
import torch
from numpy.typing import NDArray

from .mean_std_normalization import _reshape_stats
from .normalization_protocol import NormalizationProtocol


class RangeNormalization(NormalizationProtocol):
    """Normalize an image or image patch to [0, 1] range.

    This transform expects C(Z)YX dimensions for normalization and BC(Z)YX
    for denormalization. Returns float32 arrays.

    Parameters
    ----------
    input_mins : list of float
        Minimum value per channel.
    input_maxes : list of float
        Maximum value per channel.
    target_mins : list of float, optional
        Target minimum value per channel.
    target_maxes : list of float, optional
        Target maximum value per channel.

    Attributes
    ----------
    input_mins : list of float
        Minimum value per channel.
    input_maxes : list of float
        Maximum value per channel.
    target_mins : list of float, optional
        Target minimum value per channel.
    target_maxes : list of float, optional
        Target maximum value per channel.
    eps : float
        Small value to avoid division by zero.
    """

    def __init__(
        self,
        input_mins: list[float],
        input_maxes: list[float],
        target_mins: list[float] | None = None,
        target_maxes: list[float] | None = None,
    ):
        self.input_mins = input_mins
        self.input_maxes = input_maxes
        self.target_mins = target_mins
        self.target_maxes = target_maxes
        self.eps = 1e-8

    def __call__(
        self, patch: NDArray, target: NDArray | None = None
    ) -> tuple[NDArray, NDArray | None]:
        """Apply range normalization to patch and optional target.

        Parameters
        ----------
        patch : NDArray
            Patch with shape C(Z)YX.
        target : NDArray, optional
            Target patch with shape C(Z)YX.

        Returns
        -------
        tuple of NDArray
            Normalized patch and target (target can be None).
        """
        if len(self.input_mins) != patch.shape[0]:
            raise ValueError(
                f"Number of mins ({len(self.input_mins)}) and number of "
                f"channels ({patch.shape[0]} in {patch.shape}) do not match."
            )

        if len(self.input_maxes) != patch.shape[0]:
            raise ValueError(
                f"Number of max values ({len(self.input_maxes)}) and number of "
                f"channels ({patch.shape[0]} in {patch.shape}) do not match."
            )

        patch = patch.astype(np.float32)
        if target is not None:
            target = target.astype(np.float32)

        min_val = _reshape_stats(self.input_mins, patch.ndim, channel_axis=0)
        max_val = _reshape_stats(self.input_maxes, patch.ndim, channel_axis=0)

        norm_patch = (patch - min_val) / (max_val - min_val + self.eps)

        norm_target = None
        if target is not None:
            if self.target_mins is None or self.target_maxes is None:
                raise ValueError(
                    "Target mins and maxs must be provided if target is not None."
                )
            if len(self.target_mins) != target.shape[0]:
                raise ValueError(
                    f"Number of target mins ({len(self.target_mins)}) and number of "
                    f"channels ({target.shape[0]} in {target.shape}) do not match."
                )

            if len(self.target_maxes) != target.shape[0]:
                raise ValueError(
                    f"Number of target maxes ({len(self.target_maxes)}) and number "
                    f"of channels ({target.shape[0]} in {target.shape}) do not match."
                )
            target_mins = _reshape_stats(self.target_mins, target.ndim, channel_axis=0)
            target_maxes = _reshape_stats(
                self.target_maxes, target.ndim, channel_axis=0
            )
            norm_target = (target - target_mins) / (
                target_maxes - target_mins + self.eps
            )

        return norm_patch, norm_target

    def denormalize(self, patch: torch.Tensor) -> torch.Tensor:
        """Reverse the normalization operation for a batch of patches.

        Parameters
        ----------
        patch : torch.Tensor
            Normalized patch with shape BC(Z)YX.

        Returns
        -------
        torch.Tensor
            Denormalized patch.
        """
        if len(self.input_mins) != patch.shape[1]:
            raise ValueError(
                f"Number of mins ({len(self.input_mins)}) and number of "
                f"channels ({patch.shape[1]} in {patch.shape} for BC(Z)YX) "
                "do not match."
            )

        patch = patch.to(dtype=torch.float32)

        input_mins = _reshape_stats(self.input_mins, patch.ndim, channel_axis=1)
        input_maxes = _reshape_stats(self.input_maxes, patch.ndim, channel_axis=1)

        input_mins = torch.from_numpy(input_mins).to(patch.device)
        input_maxes = torch.from_numpy(input_maxes).to(patch.device)

        return patch * (input_maxes - input_mins) + input_mins



from collections.abc import Sequence

import zarr
from numpy.typing import DTypeLike, NDArray

from careamics.dataset.dataset_utils import reshape_array

from .image_utils.image_stack_utils import channel_slice, pad_patch, reshape_array_shape


class ZarrImageStack:
    """
    A class for extracting patches from an image stack that is stored as a zarr array.
    """

    def __init__(self, group: zarr.Group, data_path: str, axes: str):
        if not isinstance(group, zarr.Group):
            raise TypeError(f"group must be a zarr.Group instance, got {type(group)}.")

        self._group = group
        self._store = str(group.store_path)
        try:
            self._array = group[data_path]
        except KeyError as e:
            raise ValueError(
                f"Did not find array at '{data_path}' in store '{self._store}'."
            ) from e

        if not isinstance(self._array, zarr.Array):
            raise TypeError(
                f"data at path '{data_path}' must be a zarr.Array instance, "
                f"got {type(self._array)}."
            )

        self._source = self._array.store_path

        # TODO: validate axes
        #   - must contain XY
        #   - must be subset of STCZYX
        self._original_axes = axes
        self._original_data_shape: tuple[int, ...] = self._array.shape
        self.data_shape = reshape_array_shape(axes, self._original_data_shape)
        self._data_dtype = self._array.dtype
        self._chunk_size = reshape_array_shape(
            axes, self._array.chunks, add_singleton=False
        )
        self._shard_size = (
            reshape_array_shape(axes, self._array.shards, add_singleton=False)
            if self._array.shards is not None
            else None
        )

    # Used to identify the source of the data and write to similar path during pred
    @property
    def source(self) -> str:
        # e.g. file://data/bsd68.zarr/train/
        return str(self._source)

    @property
    def chunks(self) -> Sequence[int]:
        """Chunks size in the order of data_shape (SC(Z)YX)."""
        return self._chunk_size

    @property
    def shards(self) -> Sequence[int] | None:
        """Shard size in the order of data_shape (SC(Z)YX)."""
        return self._shard_size

    @property
    def data_dtype(self) -> DTypeLike:
        return self._data_dtype

    def extract_patch(
        self,
        sample_idx: int,
        channels: Sequence[int] | None,  # `channels = None` to select all channels,
        coords: Sequence[int],
        patch_size: Sequence[int],
    ) -> NDArray:
        # original axes assumed to be any subset of STCZYX (containing YX), in any order
        # arguments must be transformed to index data in original axes order
        # to do this: loop through original axes and append correct index/slice
        #   for each case: STCZYX
        #   Note: if any axis is not present in original_axes it is skipped.

        # guard for no S and T in original axes
        if ("S" not in self._original_axes) and ("T" not in self._original_axes):
            if sample_idx not in [0, -1]:
                raise IndexError(
                    f"Sample index {sample_idx} out of bounds for S axes with size "
                    f"{self.data_shape[0]}"
                )

        # check that channels are within bounds
        if channels is not None:
            max_channel = self.data_shape[1] - 1  # channel is second dimension
            for ch in channels:
                if ch > max_channel:
                    raise ValueError(
                        f"Channel index {ch} is out of bounds for data with "
                        f"{self.data_shape[1]} channels. Check the provided `channels` "
                        f"parameter in the configuration for erroneous channel "
                        f"indices."
                    )

        patch_slice: list[int | slice] = []
        for d in self._original_axes:
            if d == "S":
                patch_slice.append(self._get_S_index(sample_idx))
            elif d == "T":
                patch_slice.append(self._get_T_index(sample_idx))
            elif d == "C":
                patch_slice.append(channel_slice(channels))  # type: ignore
            elif d == "Z":
                patch_slice.append(slice(coords[0], coords[0] + patch_size[0]))
            elif d == "Y":
                y_idx = 0 if "Z" not in self._original_axes else 1
                patch_slice.append(
                    slice(coords[y_idx], coords[y_idx] + patch_size[y_idx])
                )
            elif d == "X":
                x_idx = 1 if "Z" not in self._original_axes else 2
                patch_slice.append(
                    slice(coords[x_idx], coords[x_idx] + patch_size[x_idx])
                )
            else:
                raise ValueError(f"Unrecognised axis '{d}', axes should be in STCZYX.")

        patch_data: NDArray = self._array[tuple(patch_slice)]  # type: ignore
        patch_axes = self._original_axes.replace("S", "").replace("T", "")
        patch_data = reshape_array(patch_data, patch_axes)[0]  # remove first sample dim
        patch = pad_patch(coords, patch_size, self.data_shape, patch_data)

        return patch

    def _get_T_index(self, sample_idx: int) -> int:
        """Get T index given `sample_idx`."""
        if "T" not in self._original_axes:
            raise ValueError("No 'T' axis specified in original data axes.")
        axis_idx = self._original_axes.index("T")
        dim = self._original_data_shape[axis_idx]

        # new S' = S*T
        # T_idx = S_idx' // T_size
        # S_idx = S_idx' % T_size
        # - floor divide finds the row
        # - modulus finds how far along the row i.e. the column
        return sample_idx % dim

    def _get_S_index(self, sample_idx: int) -> int:
        """Get S index given `sample_idx`."""
        if "S" not in self._original_axes:
            raise ValueError("No 'S' axis specified in original data axes.")
        if "T" in self._original_axes:
            T_axis_idx = self._original_axes.index("T")
            T_dim = self._original_data_shape[T_axis_idx]

            # new S' = S*T
            # T_idx = S_idx' // T_size
            # S_idx = S_idx' % T_size
            # - floor divide finds the row
            # - modulus finds how far along the row i.e. the column
            return sample_idx // T_dim
        else:
            return sample_idx



"""Normalization types."""

__all__ = [
    "MeanStdNormalization",
    "NoNormalization",
    "RangeNormalization",
    "create_normalization",
    "resolve_normalization_config",
]

from .mean_std_normalization import MeanStdNormalization
from .no_normalization import NoNormalization
from .normalization_factory import create_normalization
from .range_normalization import RangeNormalization
from .statistics import resolve_normalization_config



"""Prediction utilities for the NG Dataset."""

__all__ = [
    "combine_samples",
    "convert_prediction",
    "decollate_image_region_data",
    "stitch_prediction",
    "stitch_single_prediction",
]

from .convert_prediction import (
    combine_samples,
    convert_prediction,
    decollate_image_region_data,
)
from .stitch_prediction import stitch_prediction, stitch_single_prediction



"""Pydantic models representing coordinate and patch filters."""

__all__ = [
    "FilterConfig",
    "MaskFilterConfig",
    "MaxFilterConfig",
    "MeanSTDFilterConfig",
    "ShannonFilterConfig",
]

from .filter_config import FilterConfig
from .mask_filter_config import MaskFilterConfig
from .max_filter_config import MaxFilterConfig
from .meanstd_filter_config import MeanSTDFilterConfig
from .shannon_filter_config import ShannonFilterConfig



import pytest

from careamics.config.ng_configs import N2VConfiguration
from careamics.config.ng_factories import (
    create_advanced_n2v_config,
    create_n2v_config,
    create_structn2v_config,
)
from careamics.config.support import (
    SupportedPixelManipulation,
    SupportedStructAxis,
)


# TODO standard just calls advanced, where should test reside?
# - arguably standard should only test the parameters that it enforces, but these are
# default
# - advanced should test what it actaully does
class TestStandardConfig:
    """Test the standard N2V configuration factory."""

    def test_create_standard_config(self):
        """Test that N2V configuration can be created."""
        config = create_n2v_config(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
        )
        assert isinstance(config, N2VConfiguration)

    def test_no_aug(self):
        """Test no augmentation."""
        config = create_n2v_config(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            augmentations=[],
        )
        assert config.data_config.transforms == []

    def test_num_epochs_and_num_steps(self):
        """Test that both num_epochs and num_steps can be set simultaneously."""
        num_epochs = 25
        num_steps = 500

        config = create_n2v_config(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            num_epochs=num_epochs,
            num_steps=num_steps,
        )
        assert (
            config.training_config.lightning_trainer_config["max_epochs"] == num_epochs
        )
        assert (
            config.training_config.lightning_trainer_config["limit_train_batches"]
            == num_steps
        )

    # TODO arguably this should be tested at the level of the data config only
    def test_czi_with_T_axes(self):
        """Test that SCTYX is accepted by N2V configuration for CZI data."""
        config = create_n2v_config(
            experiment_name="test",
            data_type="czi",
            axes="SCTYX",
            patch_size=[32, 64, 64],
            batch_size=8,
            n_channels=1,
        )
        assert config.data_config.axes == "SCTYX"
        assert len(config.data_config.patching.patch_size) == 3
        assert config.algorithm_config.model.conv_dims == 3


class TestStructConfig:
    def test_n2v2_structn2v(self):
        """Test n2v2 and structn2v params are passed correctly."""
        n2v2 = True
        struct_mask_axis = SupportedStructAxis.HORIZONTAL.value
        struct_n2v_span = 15

        config = create_structn2v_config(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            struct_n2v_axis=struct_mask_axis,
            struct_n2v_span=struct_n2v_span,
            use_n2v2=n2v2,
        )
        assert (
            config.algorithm_config.n2v_config.strategy
            == SupportedPixelManipulation.MEDIAN.value
        )
        assert config.algorithm_config.n2v_config.struct_mask_axis == struct_mask_axis
        assert config.algorithm_config.n2v_config.struct_mask_span == struct_n2v_span


class TestAdvancedConfig:
    """Test the advanced N2V configuration factory"""

    @pytest.mark.parametrize(
        "axes, n_channels, channels, error",
        [
            # Valid cases
            ("YX", None, None, False),
            ("YX", 1, None, False),
            ("CYX", 1, None, False),
            ("CYX", 1, [0], False),
            ("CYX", 3, None, False),
            ("CYX", 3, [0, 2, 3], False),
            ("CYX", None, [0, 2, 3], False),
            # no channels allowed
            ("YX", None, [0], True),
            ("YX", 2, None, True),
            # mismatched channels
            ("YX", 1, [0, 2], True),
            ("YX", 3, [0, 2], True),
        ],
    )
    def test_channels(self, axes, n_channels, channels, error):
        """Test the various settings with channel parameters"""
        if error:
            with pytest.raises(ValueError):
                _ = create_advanced_n2v_config(
                    experiment_name="test",
                    data_type="tiff",
                    axes=axes,
                    patch_size=[64, 64],
                    batch_size=8,
                    n_channels=n_channels,
                    channels=channels,
                )
        else:
            config = create_advanced_n2v_config(
                experiment_name="test",
                data_type="tiff",
                axes=axes,
                patch_size=[64, 64],
                batch_size=8,
                n_channels=n_channels,
                channels=channels,
            )

            if channels is None:
                if n_channels is None:
                    n_channels = 1  # default to 1 channel

                assert config.algorithm_config.model.in_channels == n_channels
                assert config.algorithm_config.model.num_classes == n_channels
            else:
                assert config.algorithm_config.model.in_channels == len(channels)
                assert config.algorithm_config.model.num_classes == len(channels)

    def test_num_workers(self):
        """Test that num_workers can be set and overrriden by train dataloader."""
        num_workers = 4

        config: N2VConfiguration = create_advanced_n2v_config(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            num_workers=num_workers,
        )
        assert config.data_config.train_dataloader_params["num_workers"] == num_workers
        assert config.data_config.val_dataloader_params["num_workers"] == num_workers
        assert config.data_config.pred_dataloader_params["num_workers"] == num_workers

        # test overrride
        alt_num_workers = 2
        config: N2VConfiguration = create_advanced_n2v_config(
            experiment_name="test",
            data_type="tiff",
            axes="YX",
            patch_size=[64, 64],
            batch_size=8,
            num_workers=num_workers,
            train_dataloader_params={"num_workers": alt_num_workers},
        )
        assert (
            config.data_config.train_dataloader_params["num_workers"] == alt_num_workers
        )
        assert config.data_config.val_dataloader_params["num_workers"] == num_workers
        assert config.data_config.pred_dataloader_params["num_workers"] == num_workers



from collections.abc import Sequence
from pathlib import Path

import numpy as np
import pytest
import zarr
from numpy.typing import NDArray

from careamics.dataset.dataset_utils import reshape_array
from careamics.dataset_ng.image_stack import ZarrImageStack
from careamics.dataset_ng.image_stack.image_utils import channel_slice

# TODO test _reshaped_data_shape


def create_zarr(
    file_path: Path,
    data_path: str,
    data: NDArray,
    shards: Sequence[int] | None = None,
    chunks: Sequence[int] | None = None,
) -> zarr.Group:
    group = zarr.create_group(file_path.resolve())

    # create array
    array = group.create(
        name=data_path,
        shape=data.shape,
        shards=shards if shards is not None else None,
        chunks=data.shape if chunks is None else chunks,
        dtype=np.uint16,
    )
    # write data
    array[...] = data

    return group


@pytest.mark.parametrize(
    "original_axes, original_shape, sample_idx",
    [
        ("YX", (32, 48), 0),
        ("XYS", (48, 32, 3), 1),
        ("SXYC", (3, 48, 32, 2), 1),
        ("CYXT", (2, 32, 48, 3), 2),
        ("CXYTS", (2, 48, 32, 3, 2), 4),
        ("XCSYT", (48, 1, 2, 32, 3), 5),  # crazy one
    ],
)
def test_extract_patch_2D(
    tmp_path: Path,
    original_axes: str,
    original_shape: tuple[int, ...],
    sample_idx: int,
):
    # reference data to compare against, it is reshaped using careamics reshape_array
    data = np.arange(np.prod(original_shape)).reshape(original_shape)
    data_ref = reshape_array(data, original_axes)

    # save data as a zarr array to ininitialise image stack with
    file_path = tmp_path / "test_zarr.zarr"
    data_path = "image"

    # initialise ZarrImageStack
    group = create_zarr(file_path=file_path, data_path=data_path, data=data)
    image_stack = ZarrImageStack(group=group, data_path=data_path, axes=original_axes)

    # test extracted patch matches patch from reference data
    coords = (11, 4)
    patch_size = (16, 9)

    extracted_patch = image_stack.extract_patch(
        sample_idx=sample_idx, channels=None, coords=coords, patch_size=patch_size
    )  # return in SCZYX order
    patch_ref = data_ref[
        sample_idx,
        :,
        coords[0] : coords[0] + patch_size[0],
        coords[1] : coords[1] + patch_size[1],
    ]
    np.testing.assert_array_equal(extracted_patch, patch_ref)


@pytest.mark.parametrize(
    "axes, shape, channels",
    [
        ("CYX", (1, 32, 32), None),
        ("CYX", (1, 32, 32), [0]),
        ("CYX", (5, 32, 32), None),
        ("CYX", (5, 32, 32), [1, 4]),
        ("CYX", (5, 32, 32), [2]),
    ],
)
def test_extract_channels(
    tmp_path: Path,
    axes: str,
    shape: tuple[int, ...],
    channels: int,
):
    # reference data to compare against, it is reshaped using careamics reshape_array
    data = np.arange(np.prod(shape)).reshape(shape)

    # save data as a zarr array to ininitialise image stack with
    file_path = tmp_path / "test_zarr.zarr"
    data_path = "image"

    # initialise ZarrImageStack
    group = create_zarr(file_path=file_path, data_path=data_path, data=data)
    image_stack = ZarrImageStack(group=group, data_path=data_path, axes=axes)

    # extract patch
    patch = image_stack.extract_patch(
        sample_idx=0,
        channels=channels,
        coords=(0, 0),
        patch_size=(8, 8),
    )
    assert len(patch.shape) == 3  # no Z
    assert patch.shape[0] == len(channels) if channels is not None else data.shape[0]

    expected_patch = data[
        channel_slice(channels),
        0 : 0 + 8,
        0 : 0 + 8,
    ]
    np.testing.assert_array_equal(patch, expected_patch)


@pytest.mark.parametrize(
    "shape, axes, channels",
    [
        ((2, 3, 64, 64), "SCYX", [0, 4]),
        ((2, 3, 64, 64), "SCYX", [3]),
    ],
)
def test_extract_channel_error(
    tmp_path: Path,
    shape: tuple[int, ...],
    axes: str,
    channels: int,
):
    # reference data to compare against, it is reshaped using careamics reshape_array
    data = np.arange(np.prod(shape)).reshape(shape)

    # save data as a zarr array to ininitialise image stack with
    file_path = tmp_path / "test_zarr.zarr"
    data_path = "image"

    # initialise ZarrImageStack
    group = create_zarr(file_path=file_path, data_path=data_path, data=data)
    image_stack = ZarrImageStack(group=group, data_path=data_path, axes=axes)

    expected_msg = (
        f"Channel index {channels[-1]} is out of bounds for data with "
        f"{shape[1]} channels. Check the provided `channels` "
        f"parameter in the configuration for erroneous channel "
        f"indices."
    )

    with pytest.raises(ValueError, match=expected_msg):
        image_stack.extract_patch(
            sample_idx=0,
            channels=channels,
            coords=(0, 0),
            patch_size=(16, 16),
        )


def test_shards_and_chunks(tmp_path: Path):

    axes = "SCZYX"
    shape = (10, 3, 32, 64, 64)
    chunks = (1, 1, 4, 8, 8)
    shards = (1, 1, 8, 32, 16)

    data = np.arange(np.prod(shape)).reshape(shape)

    # save data as a zarr array to ininitialise image stack with
    file_path = tmp_path / "test_zarr.zarr"
    data_path = "image"

    # initialise ZarrImageStack
    group = create_zarr(
        file_path=file_path,
        data_path=data_path,
        data=data,
        shards=shards,
        chunks=chunks,
    )
    image_stack = ZarrImageStack(group=group, data_path=data_path, axes=axes)

    assert image_stack.chunks == chunks
    assert image_stack.shards == shards



"""Lightning callback for storing CAREamics configuration in checkpoints."""

from typing import Any

from pytorch_lightning import LightningModule, Trainer
from pytorch_lightning.callbacks import Callback

from careamics.config import TrainingConfig


class CareamicsCheckpointInfo(Callback):
    """
    Callback to save CAREamics configuration in Lightning checkpoints.

    This callback automatically stores CAREamics version, experiment name,
    and training configuration in the checkpoint file for reproducibility.

    Parameters
    ----------
    careamics_version : str
        Version of CAREamics used for training.
    experiment_name : str
        Name of the experiment.
    training_config : TrainingConfig
        Training configuration to store in checkpoint.

    Attributes
    ----------
    careamics_version : str
        Version of CAREamics used for training.
    experiment_name : str
        Name of the experiment.
    training_config : TrainingConfig
        Training configuration to store in checkpoint.
    """

    def __init__(
        self,
        careamics_version: str,
        experiment_name: str,
        training_config: TrainingConfig,
    ):
        """
        Initialize the callback.

        Parameters
        ----------
        careamics_version : str
            Version of CAREamics used for training.
        experiment_name : str
            Name of the experiment.
        training_config : TrainingConfig
            Training configuration to store in checkpoint.
        """
        super().__init__()
        self.careamics_version = careamics_version
        self.experiment_name = experiment_name
        self.training_config = training_config

    def on_save_checkpoint(
        self, trainer: Trainer, pl_module: LightningModule, checkpoint: dict[str, Any]
    ) -> None:
        """
        Lightning hook called when saving a checkpoint.

        Adds CAREamics configuration to the checkpoint dictionary.

        Parameters
        ----------
        trainer : Trainer
            Lightning trainer instance.
        pl_module : LightningModule
            Lightning module being trained.
        checkpoint : dict[str, Any]
            Checkpoint dictionary to modify.
        """
        checkpoint["careamics_info"] = {
            "version": self.careamics_version,
            "experiment_name": self.experiment_name,
            "training_config": self.training_config.model_dump(mode="json"),
        }



from typing import Protocol

import torch
from numpy.typing import NDArray


class NormalizationProtocol(Protocol):
    """Protocol for normalization strategies."""

    def __call__(
        self, patch: NDArray, target: NDArray | None = None
    ) -> tuple[NDArray, NDArray | None]:
        """Apply the normalization to the patch and target.

        Parameters
        ----------
        patch : NDArray
            Patch, 2D or 3D, shape C(Z)YX.
        target : NDArray, optional
            Target for the patch, by default None.

        Returns
        -------
        tuple of NDArray
            Transformed patch and target, the target can be returned as `None`.
        """
        ...

    def denormalize(self, patch: torch.Tensor) -> torch.Tensor:
        """Reverse the normalization operation for a batch of patches.

        Parameters
        ----------
        patch : torch.Tensor
            Patch, 2D or 3D, shape BC(Z)YX.
        """
        ...



"""Patching strategy factory."""

from collections.abc import Sequence

from careamics.config.data.ng_data_config import PatchingConfig
from careamics.config.support.supported_patching_strategies import (
    SupportedPatchingStrategy,
)

from .patching_strategy_protocol import PatchingStrategy
from .random_patching import FixedRandomPatchingStrategy, RandomPatchingStrategy
from .stratified_patching import StratifiedPatchingStrategy
from .tiling_strategy import TilingStrategy
from .whole_sample import WholeSamplePatchingStrategy


def create_patching_strategy(
    data_shapes: list[Sequence[int]], patching_config: PatchingConfig
) -> PatchingStrategy:
    """Factory function to create a patching strategy based on the provided config.

    Parameters
    ----------
    data_shapes : list of Sequence of int
        The shapes of the data stacks to be patched.
    patching_config: PatchingConfig
        The configuration for the desired patching strategy.

    Returns
    -------
    PatchingStrategy
        An instance of the specified patching strategy.
    """
    patch_class = None
    match patching_config.name:
        case SupportedPatchingStrategy.RANDOM:
            patch_class = RandomPatchingStrategy
        case SupportedPatchingStrategy.STRATIFIED:
            patch_class = StratifiedPatchingStrategy
        case SupportedPatchingStrategy.FIXED_RANDOM:
            patch_class = FixedRandomPatchingStrategy
        case SupportedPatchingStrategy.TILED:
            patch_class = TilingStrategy
        case SupportedPatchingStrategy.WHOLE:
            patch_class = WholeSamplePatchingStrategy
        case _:
            raise ValueError(f"Unsupported patching strategy: {patching_config.name}")

    # remove `name` to match the class signatures
    # tiling requires `tile_size` instead of `patch_size`, hence the aliasing
    return patch_class(
        data_shapes=data_shapes, **patching_config.model_dump(exclude={"name"})
    )



"""Data Pydantic configuration models."""

__all__ = [
    "DataConfig",
    "MaskFilterConfig",
    "MaxFilterConfig",
    "MeanSTDFilterConfig",
    "MeanStdConfig",
    "MinMaxConfig",
    "NGDataConfig",
    "NoNormConfig",
    "NormalizationConfig",
    "QuantileConfig",
    "RandomPatchingConfig",
    "ShannonFilterConfig",
    "TiledPatchingConfig",
    "WholePatchingConfig",
]

from .data_config import DataConfig
from .ng_data_config import NGDataConfig
from .normalization_config import (
    MeanStdConfig,
    MinMaxConfig,
    NoNormConfig,
    NormalizationConfig,
    QuantileConfig,
)
from .patch_filter import (
    MaskFilterConfig,
    MaxFilterConfig,
    MeanSTDFilterConfig,
    ShannonFilterConfig,
)
from .patching_strategies import (
    RandomPatchingConfig,
    TiledPatchingConfig,
    WholePatchingConfig,
)



"""Tests for MicroSplitDataModule data statistics handling."""

import numpy as np
import tifffile

from careamics.lightning.microsplit_data_module import (
    MicroSplitDataModule,
    get_train_val_data,
)
from careamics.lvae_training.dataset import (
    DataSplitType,
    DataType,
    MicroSplitDataConfig,
)
from careamics.lvae_training.dataset.types import TilingMode


def test_microsplit_datamodule_data_stats(tmp_path):
    rng = np.random.default_rng(42)

    # Create channel directories
    ch0_dir = tmp_path / "ch0"
    ch1_dir = tmp_path / "ch1"
    ch0_dir.mkdir()
    ch1_dir.mkdir()

    # Create random data for each channel
    for i in range(5):
        data_ch0 = rng.integers(0, 255, (64, 64)).astype(np.float32)
        data_ch1 = rng.integers(0, 255, (64, 64)).astype(np.float32)

        tifffile.imwrite(ch0_dir / f"img_{i}.tif", data_ch0)
        tifffile.imwrite(ch1_dir / f"img_{i}.tif", data_ch1)

    config = MicroSplitDataConfig(
        data_type=DataType.HTLIF24Data,
        image_size=(64, 64),
        grid_size=32,
        num_channels=2,
        batch_size=2,
        datasplit_type=DataSplitType.Train,
        multiscale_lowres_count=1,
        tiling_mode=TilingMode.ShiftBoundary,
        train_dataloader_params={},
        val_dataloader_params={},
    )

    data_module = MicroSplitDataModule(
        data_config=config,
        train_data=str(tmp_path),
        read_source_func=get_train_val_data,
    )

    assert data_module.data_stats is not None
    assert len(data_module.data_stats) == 2

    mean_dict, std_dict = data_module.data_stats
    assert isinstance(mean_dict, dict)
    assert isinstance(std_dict, dict)
    assert "input" in mean_dict
    assert "input" in std_dict



"""CAREamics Pydantic configurations."""

__all__ = [
    "CAREAlgorithm",
    "CheckpointConfig",
    "Configuration",
    "DataConfig",
    "GaussianMixtureNMConfig",
    "HDNAlgorithm",
    "InferenceConfig",
    "LVAEConfig",
    "LVAELossConfig",
    "MicroSplitAlgorithm",
    "MultiChannelNMConfig",
    "N2NAlgorithm",
    "N2VAlgorithm",
    "NGDataConfig",
    "PN2VAlgorithm",
    "TrainingConfig",
    "UNetBasedAlgorithm",
    "UNetConfig",
    "VAEBasedAlgorithm",
    "algorithm_factory",
    "create_care_configuration",
    "create_hdn_configuration",
    "create_microsplit_configuration",
    "create_n2n_configuration",
    "create_n2v_configuration",
    "create_ng_data_configuration",
    "create_pn2v_configuration",
    "load_configuration",
    "load_configuration_ng",
    "save_configuration",
]

from .algorithms import (
    CAREAlgorithm,
    HDNAlgorithm,
    MicroSplitAlgorithm,
    N2NAlgorithm,
    N2VAlgorithm,
    PN2VAlgorithm,
    UNetBasedAlgorithm,
    VAEBasedAlgorithm,
)
from .architectures import LVAEConfig, UNetConfig
from .configuration import Configuration
from .configuration_factories import (
    algorithm_factory,
    create_care_configuration,
    create_hdn_configuration,
    create_microsplit_configuration,
    create_n2n_configuration,
    create_n2v_configuration,
    create_pn2v_configuration,
)
from .data import DataConfig, NGDataConfig
from .data.inference_config import InferenceConfig
from .lightning.callbacks import CheckpointConfig
from .lightning.training_config import TrainingConfig
from .losses.loss_config import LVAELossConfig
from .ng_factories.data_factory import create_ng_data_configuration
from .noise_model import (
    GaussianMixtureNMConfig,
    MultiChannelNMConfig,
)
from .utils.configuration_io import (
    load_configuration,
    load_configuration_ng,
    save_configuration,
)



from typing import Literal

import numpy as np
import pytest
import torch
from torch.nn.functional import pad

from careamics.transforms.pixel_manipulation import (
    _apply_struct_mask,
    _get_stratified_coords,
    median_manipulate,
    uniform_manipulate,
)
from careamics.transforms.pixel_manipulation_torch import (
    _apply_struct_mask_torch,
    _create_center_pixel_mask,
    _create_struct_mask,
    _get_stratified_coords_torch,
    _get_subpatch_coords,
    median_manipulate_torch,
    uniform_manipulate_torch,
)
from careamics.transforms.struct_mask_parameters import StructMaskParameters


@pytest.mark.parametrize(
    "mask_pixel_perc, shape, num_iterations",
    [(0.4, (32, 32), 1000), (0.4, (10, 10, 10), 1000)],
)
def test_get_stratified_coords(mask_pixel_perc, shape, num_iterations):
    """Test the get_stratified_coords function.

    Ensure that the array of coordinates is randomly distributed across the
    image and that most pixels get selected.
    """
    rng = np.random.default_rng(42)

    # Define the dummy array
    array = np.zeros(shape)

    # Iterate over the number of iterations and add the coordinates. This is an MC
    # simulation to ensure that the coordinates are randomly distributed and not
    # biased towards any particular region.
    for _ in range(num_iterations):
        # Get the coordinates of the pixels to be masked
        coords = _get_stratified_coords(mask_pixel_perc, shape, rng)

        # Check that there is at least one coordinate chosen
        assert len(coords) > 0

        # Check every pair in the array of coordinates
        for coord_pair in coords:
            # Check that the coordinates are of the same shape as the patch dims
            assert len(coord_pair) == len(shape)

            # Check that the coordinates are positive values
            assert all(coord_pair) >= 0

            # Check that the coordinates are within the shape of the array
            assert [c <= s for c, s in zip(coord_pair, shape, strict=False)]

        # Add the 1 to the every coordinate location.
        array[tuple(np.array(coords).T.tolist())] += 1

    # Ensure that there's no strong pattern in the array and sufficient number of
    # pixels is masked.
    assert np.sum(array == 0) < np.sum(shape)


@pytest.mark.parametrize("shape", [(8, 8), (3, 8, 8), (8, 8, 8)])
def test_uniform_manipulate(ordered_array, shape):
    """Test the uniform_manipulate function.

    Ensures that the mask corresponds to the manipulated pixels, and that the
    manipulated pixels have a value taken from a ROI surrounding them.
    """
    rng = np.random.default_rng(42)

    # create the array
    patch = ordered_array(shape)

    # manipulate the array
    transform_patch, mask = uniform_manipulate(
        patch, mask_pixel_percentage=10, subpatch_size=5, rng=rng
    )

    # find pixels that have different values between patch and transformed patch
    diff_coords = np.array(np.where(patch != transform_patch))

    # find non-zero pixels in the mask
    mask_coords = np.array(np.where(mask == 1))

    # check that the transformed pixels correspond to the masked pixels
    assert np.array_equal(diff_coords, mask_coords)

    # for each pixel masked, check that the manipulated pixel value is within the roi
    for i in range(mask_coords.shape[-1]):
        # get coordinates
        coords = mask_coords[..., i]

        # get roi using slice in each dimension
        slices = tuple(
            [
                slice(max(0, coords[i] - 2), min(shape[i], coords[i] + 3))
                for i in range(-coords.shape[0] + 1, 0)  # range -4, -3, -2, -1
            ]
        )
        roi = patch[
            (...,) + slices
        ]  # TODO ellipsis needed bc singleton dim, might need to go away

        # TODO needs to be revisited !
        # check that the pixel value comes from the actual roi
        assert transform_patch[tuple(coords)] in roi


@pytest.mark.parametrize("shape", [(8, 8), (3, 8, 8), (8, 8, 8)])
def test_median_manipulate(ordered_array, shape):
    """Test the uniform_manipulate function.

    Ensures that the mask corresponds to the manipulated pixels, and that the
    manipulated pixels have a value taken from a ROI surrounding them.
    """
    rng = np.random.default_rng(42)

    # create the array
    patch = ordered_array(shape).astype(np.float32)

    # manipulate the array
    transform_patch, mask = median_manipulate(
        patch, subpatch_size=5, mask_pixel_percentage=10, rng=rng
    )

    # find pixels that have different values between patch and transformed patch
    diff_coords = np.array(np.where(patch != transform_patch))

    # find non-zero pixels in the mask
    mask_coords = np.array(np.where(mask == 1))

    # check that the transformed pixels correspond to the masked pixels
    assert np.array_equal(diff_coords, mask_coords)

    # for each pixel masked, check that the manipulated pixel value is within the roi
    for i in range(mask_coords.shape[-1]):
        # get coordinates
        coords = mask_coords[..., i]

        # get roi using slice in each dimension
        slices = tuple(
            [
                slice(max(0, coords[i] - 2), min(shape[i], coords[i] + 3))
                for i in range(coords.shape[0])  # range -4, -3, -2, -1
            ]
        )
        roi = patch[tuple(slices)]

        # remove value of roi center from roi
        roi = roi[roi != patch[tuple(coords)]]

        # check that the pixel value comes from the actual roi
        assert transform_patch[tuple(coords)] == np.median(roi)


@pytest.mark.parametrize(
    "coords, struct_axis, struct_span",
    [((2, 2), 1, 5), ((3, 4), 0, 5), ((9, 0), 0, 5), (((1, 2), (3, 4)), 1, 5)],
)
def test_apply_struct_mask(coords, struct_axis, struct_span):
    """Test the uniform_manipulate function.

    Ensures that the mask corresponds to the manipulated pixels, and that the
    manipulated pixels have a value taken from a ROI surrounding them.
    """
    rng = np.random.default_rng(42)

    struct_params = StructMaskParameters(axis=struct_axis, span=struct_span)

    # create array
    patch = np.arange(
        100,
    ).reshape((10, 10))

    # make a copy of the original patch for comparison
    original_patch = patch.copy()
    coords = np.array(coords)

    # expand the coords if only one roi is given
    if coords.ndim == 1:
        coords = coords[None, :]

    # manipulate the array
    transform_patch = _apply_struct_mask(
        patch,
        coords=coords,
        struct_params=struct_params,
        rng=rng,
    )
    changed_values = patch[np.where(original_patch != transform_patch)]

    # check that the transformed pixels correspond to the masked pixels
    transformed = []
    axis = 1 - struct_axis
    for i in range(coords.shape[0]):
        # get indices to mask
        indices_to_mask = [
            c
            for c in range(
                max(0, coords[i, axis] - struct_span // 2),
                min(transform_patch.shape[1], coords[i, axis] + struct_span // 2) + 1,
            )
            if c != coords[i, axis]
        ]

        # add to transform
        if struct_axis == 0:
            transformed.append(transform_patch[coords[i, 0]][indices_to_mask])
        else:
            transformed.append(transform_patch[:, coords[i, 1]][indices_to_mask])

    assert np.array_equal(
        np.sort(changed_values), np.sort(np.concatenate(transformed, axis=0))
    )


@pytest.mark.parametrize(
    "coords, struct_axis, struct_span",
    [
        ((1, 2, 2), 1, 5),
        ((2, 3, 4), 0, 5),
        ((0, 9, 0), 0, 5),
        (((2, 1, 2), (1, 9, 0), (0, 3, 4)), 1, 5),
    ],
)
def test_apply_struct_mask_3D(coords, struct_axis, struct_span):
    """Test the uniform_manipulate function.

    Ensures that the mask corresponds to the manipulated pixels, and that the
    manipulated pixels have a value taken from a ROI surrounding them.
    """
    rng = np.random.default_rng(42)

    struct_params = StructMaskParameters(axis=struct_axis, span=struct_span)

    # create array
    patch = np.arange(
        100 * 3,
    ).reshape((3, 10, 10))

    # make a copy of the original patch for comparison
    original_patch = patch.copy()
    coords = np.array(coords)

    # expand the coords if only one roi is given
    if coords.ndim == 1:
        coords = coords[None, :]

    # manipulate the array
    transform_patch = _apply_struct_mask(
        patch,
        coords=coords,
        struct_params=struct_params,
        rng=rng,
    )
    changed_values = patch[np.where(original_patch != transform_patch)]

    # check that the transformed pixels correspond to the masked pixels
    transformed = []
    axis = -2 + 1 - struct_axis
    for i in range(coords.shape[0]):
        # get indices to mask
        indices_to_mask = [
            c
            for c in range(
                max(0, coords[i, axis] - struct_span // 2),
                min(transform_patch.shape[1] - 1, coords[i, axis] + struct_span // 2)
                + 1,
            )
            if c != coords[i, axis]
        ]

        # add to transform
        if struct_axis == 0:
            transformed.append(
                transform_patch[coords[i, 0], coords[i, 1]][indices_to_mask]
            )
        else:
            transformed.append(
                transform_patch[coords[i, 0], :, coords[i, 2]][indices_to_mask]
            )

    assert np.array_equal(
        np.sort(changed_values), np.sort(np.concatenate(transformed, axis=0))
    )


@pytest.mark.parametrize(
    "mask_pixel_perc, shape, num_iterations",
    [(0.4, (4, 32, 32), 1000), (0.4, (4, 10, 10, 10), 1000)],
)
def test_get_stratified_coords_torch(mask_pixel_perc, shape, num_iterations):
    """Test the get_stratified_coords function.

    Ensure that the array of coordinates is randomly distributed across the
    image and that most pixels get selected.
    """
    rng = torch.Generator().manual_seed(42)

    # Define the dummy tensor
    tensor = torch.zeros(shape)

    # Iterate over the number of iterations and add the coordinates. This is an MC
    # simulation to ensure that the coordinates are randomly distributed and not
    # biased towards any particular region.
    for _ in range(num_iterations):
        # Get the coordinates of the pixels to be masked
        coords = _get_stratified_coords_torch(mask_pixel_perc, shape, rng)

        # Check that there is at least one coordinate chosen
        assert coords.numel() > 0

        # Check every pair in the tensor of coordinates
        for coord_pair in coords:
            # Check that the coordinates are of the same shape as the patch dims
            assert len(coord_pair) == len(shape)

            # Check that the coordinates are positive values
            assert all(coord_pair >= 0)

            # Check that the coordinates are within the shape of the tensor
            assert all(coord_pair < torch.tensor(shape))

        # Add 1 to every coordinate location
        indices = tuple(coords.T.tolist())
        tensor[indices] += 1

    # Ensure that there's no strong pattern in the tensor and sufficient number of
    # pixels are masked.
    assert torch.sum(tensor == 0) < torch.prod(torch.tensor(shape))


@pytest.mark.parametrize("shape", [(4, 8, 8), (4, 3, 8, 8), (4, 8, 8, 8)])
def test_uniform_manipulate_torch(ordered_array, shape):
    """Test the uniform_manipulate function.

    Ensures that the mask corresponds to the manipulated pixels, and that the
    manipulated pixels have a value taken from a ROI surrounding them.
    """
    rng = torch.Generator().manual_seed(42)

    # Create the tensor
    patch = torch.tensor(ordered_array(shape))

    # Manipulate the tensor
    transform_patch, mask = uniform_manipulate_torch(
        patch, mask_pixel_percentage=10, subpatch_size=5, rng=rng
    )

    # Find pixels that have different values between patch and transformed patch
    diff_coords = torch.nonzero(patch != transform_patch, as_tuple=False).T

    # Find non-zero pixels in the mask
    mask_coords = torch.nonzero(mask == 1, as_tuple=False).T

    # Check that the transformed pixels correspond to the masked pixels
    assert torch.equal(diff_coords, mask_coords)

    # For each pixel masked, check that the manipulated pixel value is within the ROI
    for i in range(mask_coords.shape[-1]):
        # get coordinates
        coords = mask_coords[..., i]

        # get roi using slice in each dimension
        slices = tuple(
            [
                slice(max(0, coords[i] - 2), min(shape[i], coords[i] + 3))
                for i in range(-coords.shape[0] + 1, 0)  # range -4, -3, -2, -1
            ]
        )
        roi = patch[
            (...,) + slices
        ]  # TODO ellipsis needed bc singleton dim, might need to go away

        # TODO needs to be revisited !
        # check that the pixel value comes from the actual roi
        assert transform_patch[tuple(coords.tolist())] in roi


@pytest.mark.parametrize("shape", [(4, 8, 8), (4, 3, 8, 8), (4, 8, 8, 8)])
@pytest.mark.parametrize("apply_struct", [False, True])
def test_median_manipulate_torch(ordered_array, shape, apply_struct: bool):
    """Test the median_manipulate function.

    Ensures that the mask corresponds to the manipulated pixels, and that the
    manipulated pixels have a value taken from a ROI surrounding them.
    """
    rng = torch.Generator().manual_seed(42)

    # Create the tensor
    patch = torch.from_numpy(ordered_array(shape)).float()
    subpatch_size = 5

    struct_params = StructMaskParameters(0, 3) if apply_struct else None
    # Manipulate the tensor
    transform_patch, mask = median_manipulate_torch(
        patch,
        subpatch_size=subpatch_size,
        mask_pixel_percentage=10,
        struct_params=struct_params,
        rng=rng,
    )

    # Find pixels that have different values between patch and transformed patch
    diff_coords = torch.nonzero(patch != transform_patch, as_tuple=False).T

    # Find non-zero pixels in the mask
    mask_coords = torch.nonzero(mask == 1, as_tuple=False).T

    # more than just the masked pixels are changed with structN2V
    if struct_params is None:
        # Check that the transformed pixels correspond to the masked pixels
        assert torch.equal(diff_coords, mask_coords)

    # For each pixel masked, check that the manipulated pixel value is within the ROI
    for i in range(mask_coords.shape[-1]):
        # Get coordinates
        coords = mask_coords[:, i]

        # Get ROI using slice in each dimension
        slices = [slice(coords[0], coords[0] + 1)] + [
            slice(max(0, coords[j] - 2), min(shape[j], coords[j] + 3))
            for j in range(1, len(coords))
        ]
        roi = patch[slices]

        ndims = len(shape) - 1
        if struct_params is not None:
            roi_mask = _create_struct_mask(
                ndims, subpatch_size, struct_params, torch.device("cpu")
            )
        else:
            roi_mask = _create_center_pixel_mask(
                ndims, subpatch_size, torch.device("cpu")
            )

        # Check that the pixel value comes from the actual ROI
        if roi.numel() == subpatch_size**2 - 1:
            roi = roi[roi_mask]
            assert torch.equal(transform_patch[tuple(coords)], torch.median(roi))


@pytest.mark.parametrize(
    "patch_shape, coords, struct_axis, struct_span",
    [
        # 2D
        ((4, 10, 10), (2, 2, 2), 1, 5),
        ((4, 10, 10), (3, 3, 4), 0, 5),
        ((4, 10, 10), (1, 9, 0), 0, 5),
        ((4, 10, 10), ((1, 1, 2), (1, 3, 4)), 1, 5),
        # 3D
        ((4, 10, 10, 10), (2, 2, 2, 2), 1, 5),
        ((4, 10, 10, 10), (3, 3, 4, 5), 0, 5),
        ((4, 10, 10, 10), (1, 9, 0, 0), 0, 5),
        ((4, 10, 10, 10), ((1, 1, 2, 3), (1, 3, 4, 5)), 1, 5),
    ],
)
def test_apply_struct_mask_torch(patch_shape, coords, struct_axis, struct_span):
    """Test the _apply_struct_mask function.

    Ensures that the mask corresponds to the manipulated pixels, and that the
    manipulated pixels have a value taken from a ROI surrounding them.
    """
    rng = torch.Generator().manual_seed(42)

    struct_params = StructMaskParameters(axis=struct_axis, span=struct_span)

    # Create tensor
    patch = (
        torch.arange(0, torch.prod(torch.tensor(patch_shape)))
        .reshape(patch_shape)
        .float()
    )

    # Make a copy of the original patch for comparison
    original_patch = patch.clone()
    coords = torch.tensor(coords, dtype=torch.int32)

    # Expand the coords if only one ROI is given
    if coords.ndim == 1:
        coords = coords.unsqueeze(0)

    # Manipulate the tensor
    transform_patch = _apply_struct_mask_torch(
        patch,
        coords=coords,
        struct_params=struct_params,
        rng=rng,
    )
    changed_values = patch[original_patch != transform_patch]

    # Check that the transformed pixels correspond to the masked pixels
    transformed = []
    axis = -1 - struct_axis

    for i in range(coords.shape[0]):
        # get indices to mask
        indices_to_mask = [
            c
            for c in range(
                max(0, coords[i, axis] - struct_span // 2),
                min(transform_patch.shape[2], coords[i, axis] + struct_span // 2) + 1,
            )
            if c != coords[i, axis]
        ]

        ndims = patch.ndim
        # axis_slice slices the patch along the struct axis, at the masked coord
        # indices_to_mask then selects the pixels in this row/col that have been masked
        axis_slice = tuple(
            coords[i, d] if d != axis else slice(None, None, None)
            for d in range(-ndims, 0, 1)
        )
        transformed.append(transform_patch[*axis_slice][indices_to_mask])

    assert torch.equal(
        torch.sort(changed_values).values, torch.sort(torch.cat(transformed)).values
    )


@pytest.mark.parametrize(
    "batch_shape, center_coords",
    [
        [
            (2, 32, 32),
            [(0, 1, 2), (0, 8, 16), (1, 17, 9), (1, 30, 31)],
        ],
        [
            (2, 32, 32, 32),
            [(0, 1, 2, 3), (0, 8, 16, 16), (1, 17, 9, 9), (1, 30, 31, 32)],
        ],
    ],
    ids=["2D", "3D"],
)
# center coords includes batch dimension
def test_get_subpatch_coords(
    batch_shape: tuple[int, ...], center_coords: list[tuple[int, ...]]
):
    """Test the subpatch coordinates match the true subpatch location."""
    # The idea of this test is to:
    # - create the full patch coordinates with torch.meshgrid
    # - slice full coords at the subpatch location, this should match the calc coords

    subpatch_size = 11
    ndims = len(batch_shape) - 1

    full_coords = torch.stack(torch.meshgrid(*[torch.arange(d) for d in batch_shape]))
    # padding with replicate mode emulates clamping behavior in implementation.
    pad_offset = subpatch_size // 2 + 1
    full_coords = pad(
        full_coords, tuple(pad_offset for _ in range(ndims * 2)), mode="replicate"
    )

    calculated_coords = _get_subpatch_coords(
        torch.tensor(center_coords), subpatch_size, batch_shape
    )
    # check each subpatch coords in a loop
    for i in range(len(center_coords)):
        # remember calculated coords shape is (D, N, S, S, ...), see function docstring
        # so use the second dimension to index each subpatch coords.
        subpatch_coords = calculated_coords[:, i]
        center_coord = center_coords[i]
        batch_index = center_coord[0]
        offset_spatial_center = (c + pad_offset for c in center_coord[1:])
        expected_slice = (
            slice(c - subpatch_size // 2, c + subpatch_size // 2 + 1)
            for c in offset_spatial_center
        )
        expected_coords = full_coords[:, batch_index, *expected_slice]

        torch.testing.assert_close(subpatch_coords, expected_coords)


@pytest.mark.parametrize("n_dims", [2, 3])
@pytest.mark.parametrize("subpatch_size", [5, 7, 11])
def test_create_center_pixel_mask(n_dims: int, subpatch_size: int):
    mask_tensor = _create_center_pixel_mask(n_dims, subpatch_size, torch.device("cpu"))
    mask = mask_tensor.detach().numpy()

    assert np.count_nonzero(~mask) == 1  # only one value masked

    centre_idx = subpatch_size // 2
    # the coordinate of the masked value should be the center index
    coord = np.where(mask == 0)
    for c in coord:
        assert c == centre_idx


@pytest.mark.parametrize("n_dims", [2, 3])
@pytest.mark.parametrize("subpatch_size", [6, 10])
def test_center_pixel_mask_even_size_error(n_dims: int, subpatch_size: int):
    """Test that even sized subpatch sizes are not allowed."""
    with pytest.raises(ValueError):
        _ = _create_center_pixel_mask(n_dims, subpatch_size, torch.device("cpu"))


@pytest.mark.parametrize("n_dims", [2, 3])
@pytest.mark.parametrize("subpatch_size", [5, 7, 11])
@pytest.mark.parametrize("span", [3, 5])
@pytest.mark.parametrize("axis", [0, 1])
def test_create_struct_mask(
    n_dims: int, subpatch_size: int, span: int, axis: Literal[0, 1]
):
    struct_params = StructMaskParameters(axis, span)
    mask_tensor = _create_struct_mask(
        n_dims, subpatch_size, struct_params, torch.device("cpu")
    )
    mask = mask_tensor.detach().numpy()

    if n_dims == 2:
        if axis == 0:
            expected_mask = _horizontal_struct_2D(subpatch_size, span)
        elif axis == 1:
            expected_mask = _vertical_struct_2D(subpatch_size, span)
        else:
            raise ValueError
    elif n_dims == 3:
        if axis == 0:
            expected_mask = _horizontal_struct_3D(subpatch_size, span)
        elif axis == 1:
            expected_mask = _vertical_struct_3D(subpatch_size, span)
        else:
            raise ValueError
    else:
        raise ValueError

    np.testing.assert_equal(mask, expected_mask)


@pytest.mark.parametrize("n_dims", [2, 3])
@pytest.mark.parametrize("subpatch_size", [6, 10])
def test_struct_mask_even_size_error(n_dims: int, subpatch_size: int):
    """Test that even sized subpatch sizes are not allowed."""
    struct_params = StructMaskParameters(0, 5)
    with pytest.raises(ValueError):
        _ = _create_struct_mask(
            n_dims, subpatch_size, struct_params, torch.device("cpu")
        )


# alternative implementation for testing for struct mask creation
# 2D, 3D, horizontal, vertical


def _horizontal_struct_2D(subpatch_size: int, span: int) -> np.ndarray:
    mask = np.ones((subpatch_size, subpatch_size), dtype=bool)
    center_idx = subpatch_size // 2
    span_start = center_idx - span // 2
    span_end = center_idx + span // 2 + 1
    mask[center_idx, span_start:span_end] = False
    return mask


def _vertical_struct_2D(subpatch_size: int, span: int) -> np.ndarray:
    mask = np.ones((subpatch_size, subpatch_size), dtype=bool)
    center_idx = subpatch_size // 2
    span_start = center_idx - span // 2
    span_end = center_idx + span // 2 + 1
    mask[span_start:span_end, center_idx] = False
    return mask


def _horizontal_struct_3D(subpatch_size: int, span: int) -> np.ndarray:
    mask = np.ones((subpatch_size, subpatch_size, subpatch_size), dtype=bool)
    mask_2d = _horizontal_struct_2D(subpatch_size, span)
    center_idx = subpatch_size // 2
    mask[center_idx] = mask_2d
    return mask


def _vertical_struct_3D(subpatch_size: int, span: int) -> np.ndarray:
    mask = np.ones((subpatch_size, subpatch_size, subpatch_size), dtype=bool)
    mask_2d = _vertical_struct_2D(subpatch_size, span)
    center_idx = subpatch_size // 2
    mask[center_idx] = mask_2d
    return mask
